/*
 * Copyright 2008-2011 Femke Ongenae, Stijn Verstichel
 * distributed under the terms of the AGPL License
 * 
 * This file is part of OTAGen.
 * 
 * OTAGen is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Femke Ongenae & Stijn Verstichel
 * Ghent University - INTEC - IBCN
 * Zuiderpoort Block C0 bus 201
 * Gaston Crommenlaan 8
 * B-9050 Ghent
 * Belgium
 * E-mail: Femke.Ongenae@intec.ugent.be
 * E-mail: Stijn.Verstichel@intec.ugent.be
 * 
 */
package otagen;

import com.hp.hpl.jena.ontology.AllDifferent;
import com.hp.hpl.jena.ontology.EnumeratedClass;
import com.hp.hpl.jena.ontology.Individual;
import com.hp.hpl.jena.ontology.ObjectProperty;
import com.hp.hpl.jena.ontology.OntClass;
import com.hp.hpl.jena.ontology.OntModel;
import com.hp.hpl.jena.ontology.OntProperty;
import com.hp.hpl.jena.ontology.OntResource;
import com.hp.hpl.jena.ontology.UnionClass;
import com.hp.hpl.jena.rdf.model.Literal;
import com.hp.hpl.jena.rdf.model.Property;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.StmtIterator;
import com.hp.hpl.jena.util.iterator.ExtendedIterator;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Random;
import otagen.datatypes.Cluster;
import otagen.datatypes.DeferredClusterOntClass;
import otagen.datatypes.DeferredInterClusterOntProperty;
import otagen.datatypes.IndividualGeneratorSessionVariables;
import otagen.datatypes.OntoList;
import otagen.feedback.AboxInformation;

/**
 * Generates the individuals/instances that conform to the ontology.
 * 
 * @author Femke Ongenae
 * @author Stijn Verstichel
 */
public class AboxGen {

    private boolean debug = false;
    private static int ccc = 0;
    //The base URI of the ontology
    private static String baseURI = "http://localhost/owl/ontologies/generated#";
    //a pseudo random generator (with seed);
    private Random pseudoRand = null;
    //holds the information about the generated A-Box.
    private AboxInformation feedBack = null;
    //Specifies the percentage of fully connected instances amongst the 
    //instances of the concepts in the cluster
    private int clusterConnectivity = 0;
    //Specifies the amount of clusters to be generated in the Ontology Model
    private int clusterCount = 0;
    //Specifies the minimum amount of concepts to be included in the cluster
    private int minClusterSize = 0;
    //Specifies the maximum amount of concept to be included in a cluster
    private int maxClusterSize = 0;
    //Specifies the minimum amount of individuals to be generated
    private int minIndividualCount = 0;
    //indicates how many instances of the datatypeproperty are to be generated in the case the property is non-functional
    private int datatypePropertyCount = 0;
    //indicates how many of the individuals should have their datatype properties filled in
    private int individualDatatypePropertyPercentage = 0;
    //Indicates minimum instances of the objectproperty are to be generated in the case the property is non-functional (intra-cluster)
    private int minObjectPropertyCount = 0;
    //Indicates maximum instances of the objectproperty are to be generated in the case the property is non-functional (intra-cluster)
    private int maxObjectPropertyCount = 0;
    //Holds the Jena OntModel
    private OntModel model = null;
    //Lists to hol references to concepts to be instantiated because
    //at the first processing domain and range were not set!
    private List<DeferredClusterOntClass> deferredClusterClasses = null;
    private List<DeferredInterClusterOntProperty> deferredInterclusterProperties = null;
    //Boolean Frequency-table to keep track of already instantiated property-trees in case of functional properties
    private boolean[][] instantiatedFunctionals = null;

    /**
     * Generates the instance level (A-Box) of the ontology
     * 
     * @param int clusterConnectivity: Specifies the percentage of fully 
     * connected instances amongst the instances of the concepts in the cluster
     * @param int clusterCount: Specifies the amount of clusters to be generated 
     * in the Ontology Model
     * @param int minClusterSize: Specifies the minimum amount of concepts to be 
     * included in the cluster
     * @param int maxClusterSize: Specifies the maximum amount of concept to be 
     * included in a cluster
     * @param int minIndividualCount: Specifies the minimum amount of 
     * individuals to be generated
     * @param OntModel ontModel: The Jena internal OntModel, as constructed with
     * the generated T-Box
     * @param Random pseudoRand: The seed-based random generator, the same as
     * being used by the T-Box generator.
     * @param int datatypePropertyCount: indicates how many instances of the datatypeproperty are to be generated in the case the property is non-functional
     * @param int individualDatatypePropertyPercentage: indicates how many of the individuals should have their datatype properties filled in
     * @param int minObjectPropertyCount: Indicates minimum instances of the objectproperty are to be generated in the case the property is non-functional (intra-cluster)
     * @param int maxObjectPropertyCount: Indicates maximum instances of the objectproperty are to be generated in the case the property is non-functional (intra-cluster)
     */
    public AboxGen(int clusterConnectivity, int clusterCount, int maxClusterSize,
            int minClusterSize, int minIndividualCount, OntModel model,
            Random pseudoRand, int datatypePropertyCount, int individualDatatypePropertyPercentage,
            int minObjectPropertyCount, int maxObjectPropertyCount) {
        //initialize the fields
        this.clusterConnectivity = clusterConnectivity;
        this.clusterCount = clusterCount;
        this.minClusterSize = minClusterSize;
        this.maxClusterSize = maxClusterSize;
        this.minIndividualCount = minIndividualCount;
        this.datatypePropertyCount = datatypePropertyCount;
        this.individualDatatypePropertyPercentage = individualDatatypePropertyPercentage;
        this.minObjectPropertyCount = minObjectPropertyCount;
        this.maxObjectPropertyCount = maxObjectPropertyCount;
        this.model = model;
        this.pseudoRand = pseudoRand;
        deferredClusterClasses = new ArrayList<DeferredClusterOntClass>();
        deferredInterclusterProperties = new ArrayList<DeferredInterClusterOntProperty>();

        //create an empty Aboxinformation object that will hold the feedback 
        //about this generated A-Box
        feedBack = new AboxInformation();

        //generate clusters, i.e. define the concepts that make up a cluster
        this.defineClusters();

        //generate individuals until all concepts have at least the minimal
        //amount of individuals as defined in the otagen.properties file.

        this.generateIndividuals();
        this.setAllDifferent();

        this.instantiatedFunctionals = new boolean[model.listIndividuals().toList().size()][model.listObjectProperties().toList().size()];
        for (int i = 0; i < this.instantiatedFunctionals.length; i++) {
            for (int j = 0; j < this.instantiatedFunctionals[i].length; j++) {
                this.instantiatedFunctionals[i][j] = false;
            }
        }

        //Link the minimum clusterconnectivity amount of individuals of the 
        //cluster together, using the relationships defined in the T-Box
        //between these concepts
        this.linkClusters();

        //Create inter-cluster linking by means of relationship that have a 
        //domain and range in different clusters. The minimum connectivity
        //for this is the complement of the clusterconnectivity
        this.interlinkClusters();

        //Generates values for the datatypeProperties defined in the T-Box
        this.generateDatatypePropertyInstances();

        System.out.println("[INFO] ClusterEnum: " + deferredClusterClasses);
        System.out.println("[INFO] InterClusterEnum: " + deferredInterclusterProperties);



    }

    /**
     * This method defines the cluster and their contributing concepts. The size 
     * if the clusters is randomly (Seed-based) choosen conforming the limits
     * set in the otagen.properties file.
     */
    private void defineClusters() {
        int[] clusterSizes = new int[clusterCount];
        for (int i = 0; i < clusterSizes.length; i++) {
            clusterSizes[i] = pseudoRand.nextInt(maxClusterSize - minClusterSize)
                    + minClusterSize;
            if (debug) {
                System.out.println(clusterSizes[i]);
            }
        }
        if (debug) {
            System.out.println("[DEFINECLUSTERS][DEBUG] Clustersize: "
                    + clusterSizes);
        }

        Cluster cluster = null;
        int nextConceptIndex = 0;
        ExtendedIterator eit = model.listNamedClasses();
        OntoList classes = new OntoList();
        classes.addAll(eit.toList());
        List<Cluster> clusters = new ArrayList<Cluster>();
        for (int j = 0; j < clusterCount; j++) {
            cluster = new Cluster();
            nextConceptIndex = 0;
            //While not sufficient ontclasses have been selected as 
            //cluster-members, add more ontclasses
            while (cluster.getCluster().size() < clusterSizes[j]) {
                nextConceptIndex = pseudoRand.nextInt(classes.size());
                cluster.getCluster().add(classes.getOntClass(nextConceptIndex));
            }
            clusters.add(cluster);
        }
        feedBack.setGeneratedClusters(clusters);
        if (debug) {
            System.out.println("[DEFINECLUSTERS][DEBUG] Defined clusters: "
                    + clusters);
        }
    }

    /**
     * generate individuals until all concepts have at least the minimal
     * amount of individuals as defined in the otagen.properties file.
     */
    private void generateIndividuals() {
        ExtendedIterator eit = model.listNamedClasses();
        List<OntClass> classes = eit.toList();
        classes.removeAll(model.listComplementClasses().toList());
        classes.removeAll(model.listEnumeratedClasses().toList());
        classes.removeAll(model.listRestrictions().toList());
        classes.removeAll(model.listUnionClasses().toList());
        classes.removeAll(model.listIntersectionClasses().toList());
        IndividualGeneratorSessionVariables session = new IndividualGeneratorSessionVariables(minIndividualCount,
                pseudoRand, classes);
        OntClass nextOntClass = session.getNextOntClass();
        int count = 0;
        //For all OntClasses, generate Individuals in a random order,
        //until the stopping test has been fulfilled.
        while (nextOntClass != null) {
            nextOntClass.createIndividual(baseURI + "Individual" + count);
            count++;
            nextOntClass = session.getNextOntClass();
        }
        feedBack.getIndividualCount().setGeneratedIndividuals(
                session.getGeneratedCountMap());
        if (debug) {
            System.out.println("[GENERATEINDIVIDUALS][DEBUG] Generated "
                    + "Individuals: " + feedBack.getIndividualCount().
                    getGeneratedIndividuals());
        }
    }

    /**
     * To avoid misreasoning, the indiviudals have to be defined to be
     * all defifferent. Otherwise, if no clear reason is modelled,
     * the reasoner might conclude that some individuals are actually
     * designating the same real-world instance.
     */
    private void setAllDifferent() {
        ExtendedIterator it = model.listNamedClasses();
        OntClass ontClass = null;
        AllDifferent allDifferent = null;
        allDifferent = model.createAllDifferent();
        while (it.hasNext()) {
            ontClass = (OntClass) it.next();
            //allDifferent = model.createAllDifferent();
            allDifferent.addDistinctMembers(ontClass.listInstances().toList().iterator());
        }
    }

    /**
     * Link the minimum clusterconnectivity amount of individuals of the 
     * cluster together, using the relationships defined in the T-Box
     * between these concepts
     */
    private void linkClusters() {
        int individualConnectionCount = 0;
        Cluster cluster = null;
        List<Cluster> clusters = feedBack.getGeneratedClusters();
        Iterator<Cluster> it = clusters.iterator();
        //for every cluster
        while (it.hasNext()) {
            cluster = it.next();
            individualConnectionCount = this.defineMinimumIndividualConnectionCount(cluster);
            this.linkCluster(cluster, individualConnectionCount);
        }
    }

    /**
     * This method that the minimum cluster connectivity of the concept
     * with the largest amount of individuals, and returns it for further usage
     * in the linking process.
     * 
     * @return int MinimumIndividualConnectionCount
     */
    private int defineMinimumIndividualConnectionCount(Cluster cluster) {
        OntClass ontClass = null;
        int maxIndividualCount = 0;
        Iterator<OntClass> oit = cluster.getCluster().iterator();
        while (oit.hasNext()) {
            ontClass = oit.next();
            if (ontClass.listInstances().toList().size() > maxIndividualCount) {
                maxIndividualCount = ontClass.listInstances().toList().size();
            }
        }
        maxIndividualCount *= clusterConnectivity;
        maxIndividualCount /= 100;
        // To count for rounding mistakes when casting double to int.
        maxIndividualCount++;
        if (debug) {
            System.out.println("[DEFINEMINIMUMINDIVIDUALCONNECTIONCOUNT][DEBUG] " + "MinimumIndividualConnectionCount: " + maxIndividualCount);
        }
        return maxIndividualCount;
    }

    /**
     * This method that the minimum cluster connectivity of the concept
     * with the largest amount of individuals, and returns it for further usage
     * in the linking process.
     * 
     * @return int MinimumIndividualConnectionCount
     */
    private int defineInterClusterMinimumIndividualConnectionCount() {
        OntClass ontClass = null;
        int maxIndividualCount = 0;
        ExtendedIterator oit = model.listNamedClasses();
        OntoList ontoList = new OntoList();
        ontoList.addAll(oit.toList());
        //while (oit.hasNext()) {
        for (int i = 0; i < ontoList.size(); i++) {
            ontClass = (OntClass) ontoList.getOntClass(i);
            if (ontClass.listInstances().toList().size() > maxIndividualCount) {
                maxIndividualCount = ontClass.listInstances().toList().size();
            }
        }
        maxIndividualCount *= (100 - clusterConnectivity);
        maxIndividualCount /= 100;
        // To count for rounding mistakes when casting double to int.
        maxIndividualCount++;
        if (debug) {
            System.out.println("[DEFINEMINIMUMINTERCLUSTERINDIVIDUALCONNECTIONCOUNT][DEBUG] " + "MinimumInterClusterIndividualConnectionCount: " + maxIndividualCount);
        }
        return maxIndividualCount;
    }

    /**
     * For a given cluster and given minimum amount of individuals to be 
     * included, this method connects the needed individuals with the 
     * individuals of the other concepts of the cluster, according to the
     * relations defined.
     * @param cluster
     * @param individualConnectionCount
     */
    private void linkCluster(Cluster cluster, int individualConnectionCount) {
        if (debug) {
            System.out.println("[LINKCLUSTER][DEBUG] "
                    + "Start Clustering: " + cluster.getCluster()
                    + " Connectivity: " + individualConnectionCount);
        }
        OntClass ontClass = null;
        Iterator<OntClass> it = cluster.getCluster().iterator();
        while (it.hasNext()) {
            ontClass = it.next();
            if (ontClass.isEnumeratedClass()) {
                //Save the linkage of these enumerated classed for
                //later processing as at this stage the domain and range are
                //not yet defined.
                DeferredClusterOntClass deferredOntClass = new DeferredClusterOntClass();
                deferredOntClass.setOntClass(ontClass);
                deferredOntClass.setIndividualConnectionCount(individualConnectionCount);
                deferredOntClass.setCluster(cluster);
                this.deferredClusterClasses.add(deferredOntClass);
            } else {
                this.linkOntClass(ontClass, individualConnectionCount, cluster);
            }
        }
    }

    /**
     * Starting from the given ontClass, connect as much individuals
     * with other inidivuals according to the T-Box relationships.
     * 
     * @param ontClass
     * @param individualConnectionCount
     */
    private void linkOntClass(OntClass originalOntClass, int individualConnectionCount, Cluster cluster) {
        ExtendedIterator eit = null;
        OntProperty ontProperty = null;
        if (debug) {
            System.out.println("[LINKCLUSTER][DEBUG] "
                    + "Doing OntClass: " + originalOntClass);
        }
        OntClass ontClass = model.getOntClass(originalOntClass.getURI());
        eit = model.listObjectProperties();
        OntoList ontoList = new OntoList();
        OntProperty property = null;
        while (eit.hasNext()) {
            property = (OntProperty) eit.next();
            if (property.getDomain() != null && ontClass.getURI().equalsIgnoreCase(property.getDomain().getURI())) {
                ontoList.add(property);
            }
        }

        //Do the same for the subclasses of this class as they
        //inherit the properties of the parent concept
        for (int i = 0; i < ontoList.size(); i++) {
            ontProperty = ontoList.getOntProperty(i);
            Iterator<OntClass> sit = this.getSubClasses(ontClass).iterator();
            OntClass subClass = null;
            while (sit.hasNext()) {
                subClass = sit.next();
                this.linkOntProperty(subClass, ontProperty, individualConnectionCount, cluster);
            }
        }
    }

    /**
     * Given the ontClass and ontProperty references, connect as many 
     * individuals as given in individualConnectionCount to other individuals
     * of the OntClass representing the range of the ontProperty concerned.
     * @param ontClass
     * @param ontProperty
     * @param individualConnectionCount
     */
    private void linkOntProperty(OntClass ontClass, OntProperty ontProperty,
            int individualConnectionCount, Cluster cluster) {
        OntResource range = null;
        ExtendedIterator iit = null;
        Individual individual = null;
        int actualCount = 0;

        if (debug) {
            System.out.println("[LINKCLUSTER][DEBUG] "
                    + "Doing OntProperty: " + ontProperty);
        }
        range = ontProperty.getRange();
        if (debug) {
            System.out.println("[LINKCLUSTER][DEBUG] "
                    + "Range: " + range);
        }
        if (range != null) {
            iit = ontClass.listInstances();
            OntoList ontoList = new OntoList();
            ontoList.addAll(iit.toList());
            for (int i = 0; i < ontoList.size(); i++) {
                individual = ontoList.getIndividual(i);
                this.linkIndividual(ontClass, ontProperty, range, individual, cluster);
                actualCount++;
            }
            actualCount = 0;
        } else {
            //To be checked later when the second iteration has defined domain
            //and ranges
            DeferredClusterOntClass deferredOntClass = new DeferredClusterOntClass();
            deferredOntClass.setOntClass(ontClass);
            deferredOntClass.setIndividualConnectionCount(individualConnectionCount);
            deferredOntClass.setCluster(cluster);
            this.deferredClusterClasses.add(deferredOntClass);
        }
    }

    /**
     * Given the domain individuals, the ontProperty and the rangeClass, choose
     * an individual from the range and connect it to one another.
     * @param ontProperty
     * @param range
     * @param individual
     */
    private void linkIndividual(OntClass domain, OntProperty ontProperty,
            OntResource range, Individual individual, Cluster cluster) {

        OntClass rangeClass = null;
        Individual rangeIndividual = null;
        UnionClass unionRangeClass = null;
        EnumeratedClass enumeratedClass = null;
        ExtendedIterator eit = null;
        int index = 0;
        OntClass rangeClassOperand = null;

        if (debug) {
            System.out.println("[LINKCLUSTER][DEBUG] "
                    + "Domain Individual: " + individual);
        }
        //Concrete ontClass
        if (range.isURIResource()) {
            rangeClass = model.getOntClass(range.getURI());
            OntoList rangeList = new OntoList();
            rangeList.add(rangeClass);
            rangeList.addAll(rangeClass.listSubClasses().toList());
            for (int j = 0; j < rangeList.size(); j++) {
                rangeClass = rangeList.getOntClass(j);
                if (debug) {
                    System.out.println("[LINKCLUSTER][DEBUG] "
                            + "Checking Cluster Range: " + rangeClass + " "
                            + "Cluster: " + cluster + " "
                            + (cluster.getCluster().contains(rangeClass) && cluster.getCluster().contains(domain)));
                }
                if (cluster.getCluster().contains(rangeClass) && cluster.getCluster().contains(domain)) {
                    OntoList ontoList = new OntoList();
                    ontoList.addAll(rangeClass.listInstances().toList());
                    int linkCount = pseudoRand.nextInt(this.maxObjectPropertyCount - this.minObjectPropertyCount);
                    linkCount += this.minObjectPropertyCount;
                    for (int jj = 0; jj < linkCount; jj++) {
                        rangeIndividual = ontoList.getIndividual(pseudoRand.nextInt(ontoList.size()));
                        this.doLinking(ontProperty, individual, rangeIndividual, domain, rangeClass);
                    }
                } else {
                    this.linkInterClusterSubClasses(ontProperty, domain, range, this.defineInterClusterMinimumIndividualConnectionCount());
                }
            }
            //Otherwise
        } else {
            rangeClass = range.asClass();
            //UnionClass
            if (rangeClass.isUnionClass()) {
                unionRangeClass = rangeClass.asUnionClass();
                eit = unionRangeClass.listOperands();
                index = pseudoRand.nextInt(eit.toList().size());
                OntoList ontoList = new OntoList();
                ontoList.addAll(eit.toList());
                rangeClassOperand = ontoList.getOntClass(index);
                if (cluster.getCluster().contains(rangeClass) && cluster.getCluster().contains(domain)) {
                    OntoList ontoList2 = new OntoList();
                    ontoList2.addAll(rangeClassOperand.listInstances().toList());
                    int linkCount = pseudoRand.nextInt(this.maxObjectPropertyCount - this.minObjectPropertyCount);
                    linkCount += this.minObjectPropertyCount;
                    for (int jj = 0; jj < linkCount; jj++) {
                        rangeIndividual = ontoList.getIndividual(pseudoRand.nextInt(ontoList.size()));
                        this.doLinking(ontProperty, individual, rangeIndividual, domain, rangeClass);
                    }
                    //Again do the same for the subclasses
                } else {
                    this.linkInterClusterSubClasses(ontProperty, domain, rangeClass, this.defineInterClusterMinimumIndividualConnectionCount());
                }
                //Enumerated Class
            } else if (rangeClass.isEnumeratedClass()) {
                enumeratedClass = rangeClass.asEnumeratedClass();
                eit = enumeratedClass.listOneOf();
                OntoList ontoList = new OntoList();
                ontoList.addAll(eit.toList());
                index = pseudoRand.nextInt(eit.toList().size());
                rangeClassOperand = ontoList.getOntClass(index);
                if (cluster.getCluster().contains(rangeClass) && cluster.getCluster().contains(domain)) {
                    OntoList ontoList2 = new OntoList();
                    ontoList2.addAll(rangeClassOperand.listInstances().toList());
                    int linkCount = pseudoRand.nextInt(this.maxObjectPropertyCount - this.minObjectPropertyCount);
                    linkCount += this.minObjectPropertyCount;
                    for (int jj = 0; jj < linkCount; jj++) {
                        rangeIndividual = ontoList.getIndividual(pseudoRand.nextInt(ontoList.size()));
                        this.doLinking(ontProperty, individual, rangeIndividual, domain, rangeClass);
                    }
                    //Again do the same for the subclasses
                } else {
                    this.linkInterClusterSubClasses(ontProperty, domain, rangeClass, this.defineInterClusterMinimumIndividualConnectionCount());
                }
            }
        }
    }

    /**
     * Create inter-cluster linking by means of relationship that have a 
     * domain and range in different clusters. The minimum connectivity
     * for this is the complement of the clusterconnectivity
     */
    private void interlinkClusters() {
        //get all properties and keep those linking classes from different
        //clusters
        List<OntProperty> interClusterProperties = this.getInterClusterProperties();
        //System.out.println(interClusterProperties);
        if (debug) {
            System.out.println("[INTERLINKCLUSTERS][DEBUG] "
                    + "InterCluster properties: " + interClusterProperties);
        }

        int minimumInterClusterIndividualCount = this.defineInterClusterMinimumIndividualConnectionCount();
        //Also set the minimumCount for the deffered Propertie for later
        //processing
        Iterator<DeferredInterClusterOntProperty> itt = this.deferredInterclusterProperties.iterator();
        while (itt.hasNext()) {
            itt.next().setIndividualConnectionCount(minIndividualCount);
        }

        if (debug) {
            System.out.println("[INTERLINKCLUSTERS][DEBUG] "
                    + "Individual Count: " + minimumInterClusterIndividualCount);
        }
        //In an ArrayList the sequence is deterministic according to addition
        //of the elements. No further sorting is needed!
        Iterator<OntProperty> it = interClusterProperties.iterator();
        OntProperty property = null;
        while (it.hasNext()) {
            property = it.next();
            if (debug) {
                System.out.println("[INTERLINKCLUSTERS][DEBUG] "
                        + "Doing Property: " + property);
            }
            this.linkInterCluster(property, minimumInterClusterIndividualCount);
        }

    }

    /**
     * Takes random classes from the domain and range in case that domain
     * and range are complex concepts, or the domain and the range itself in the
     * simple case. From the individuals belonging to this concepts, a random
     * individual is then taken from both domain and range and linked together
     * using the property until enough relations are established as indicated
     * by the parameter.
     * 
     * @param property
     * @param count
     */
    private void linkInterCluster(OntProperty property, int count) {
        OntResource domain = property.getDomain();
        OntResource range = property.getRange();
        if (domain != null && range != null) {
            int c = 0;
            while (c < count) {
                this.linkInterCluster(property, domain, range);
                c++;
            }
            //Also link the subconcepts with the property concerned (inheritance)
            this.linkInterClusterSubClasses(property, domain, range, count);
        } else {
            //To be stored for the later processing when domain and range will
            //have been set
            DeferredInterClusterOntProperty deferrredProperty = new DeferredInterClusterOntProperty();
            deferrredProperty.setOntProperty(property);
            deferrredProperty.setIndividualConnectionCount(minIndividualCount);
            deferredInterclusterProperties.add(deferrredProperty);
        }
    }

    /**
     * This methods used the parameters provided to also put in place
     * properties for the subconcepts of the main parent concept for which
     * the property has been defined.
     * 
     * @param property
     * @param domain
     * @param range
     * @param count
     */
    private void linkInterClusterSubClasses(OntProperty property, OntResource domain,
            OntResource range, int count) {
        if (debug) {
            System.out.println("[LINKINTERCLUSTERSUBCLASSES][DEBUG] Doing Subclasses:"
                    + property + " " + domain + " " + range);
        }

        //As plain OntClass
        if (domain.isURIResource()) {
            OntClass ontClass = model.getOntClass(domain.getURI());
            this.getInterClusterSubClass(ontClass, range, property, count);
        } else {
            //As Enumerated Class
            if (domain.asClass().isEnumeratedClass()) {
                EnumeratedClass eClass = domain.asClass().asEnumeratedClass();
                OntoList ontoList = new OntoList();
                ontoList.addAll(eClass.listOneOf().toList());
                OntClass ontClass = ontoList.getOntClass(pseudoRand.nextInt(ontoList.size()));
                this.getInterClusterSubClass(ontClass, range, property, count);
                //As Union Class
            } else if (domain.asClass().isUnionClass()) {
                UnionClass uClass = domain.asClass().asUnionClass();
                OntoList ontoList = new OntoList();
                ontoList.addAll(uClass.listOperands().toList());
                OntClass ontClass = ontoList.getOntClass(pseudoRand.nextInt(ontoList.size()));
                this.getInterClusterSubClass(ontClass, range, property, count);
            }

        }
    }

    /**
     * Due to the inability of Jena to retrieve all SubClasses wihtout an
     * attached reasoner, the list has to be defined by proprietary, and 
     * iterating code
     * 
     * Additionally this method will put in place the property between the
     * individuals of these subclasses.
     * 
     * @param domain
     * @param range
     * @param property
     * @param count
     */
    private void getInterClusterSubClass(OntClass domain, OntResource range, OntProperty property,
            int count) {
        if (domain.listSubClasses().toList().size() > 0) {
            ExtendedIterator it = domain.listSubClasses();
            OntoList ontoList = new OntoList();
            ontoList.addAll(it.toList());
            for (int i = 0; i
                    < ontoList.size(); i++) {
                this.getInterClusterSubClass(ontoList.getOntClass(i), range, property, count);
            }

            int c = 0;
            while (c < count) {
                this.linkInterCluster(property, domain, range);
                c++;

            }

        } else {
            int c = 0;
            while (c < count) {
                this.linkInterCluster(property, domain, range);
                c++;

            }
        }
    }

    /**
     * Takes a random individual belonging to the domnain and the range of
     * the supplied property and links it together.
     * 
     * @param property
     * @param domain
     * @param range
     */
    private void linkInterCluster(OntProperty property, OntResource domain,
            OntResource range) {
        //Plain real OntClass
        if (domain.isURIResource()) {
            this.linkInterCluster(property, model.getOntClass(domain.getURI()), range);
        } else {
            //As Enumerated Class
            if (domain.asClass().isEnumeratedClass()) {
                EnumeratedClass eClass = domain.asClass().asEnumeratedClass();
                OntoList ontoList = new OntoList();
                ontoList.addAll(eClass.listOneOf().toList());
                OntClass ontClass = ontoList.getOntClass(pseudoRand.nextInt(ontoList.size()));
                this.linkInterCluster(property, ontClass, range);
                //As Union Classs
            } else if (domain.asClass().isUnionClass()) {
                UnionClass uClass = domain.asClass().asUnionClass();
                OntoList ontoList = new OntoList();
                ontoList.addAll(uClass.listOperands().toList());
                OntClass ontClass = ontoList.getOntClass(pseudoRand.nextInt(ontoList.size()));
                this.linkInterCluster(property, ontClass, range);
            }
        }
    }

    /**
     * Takes a random individual belonging to the domnain and the range of
     * the supplied property and links it together.
     * 
     * @param property
     * @param domain
     * @param range
     */
    private void linkInterCluster(OntProperty property, OntClass domain,
            OntResource range) {
        //As plain Ont Class (URI)
        if (range.isURIResource()) {
            OntClass rangeClass = model.getOntClass(range.getURI());
            this.linkInterCluster(property, domain, rangeClass);
        } else {
            //As Enumerated Class
            if (range.asClass().isEnumeratedClass()) {
                EnumeratedClass eClass = range.asClass().asEnumeratedClass();
                OntoList ontoList = new OntoList();
                ontoList.addAll(eClass.listOneOf().toList());
                OntClass rangeClass = ontoList.getOntClass(pseudoRand.nextInt(ontoList.size()));
                this.linkInterCluster(property, domain, rangeClass);
                //As Union Class
            } else if (range.asClass().isUnionClass()) {
                UnionClass uClass = range.asClass().asUnionClass();
                OntoList ontoList = new OntoList();
                ontoList.addAll(uClass.listOperands().toList());
                OntClass rangeClass = ontoList.getOntClass(pseudoRand.nextInt(ontoList.size()));
                this.linkInterCluster(property, domain, rangeClass);
            }
        }
    }

    /**
     * This method is used to see if a statement has already been added
     * to the model, between the properties given. This is used by methods
     * to check whether functional properties are not violated, even in a hidden form
     *
     * @param domainIndividual
     * @param properties
     * @param rangeIndividual
     * @return
     */
    private boolean existStatement(Individual domainIndividual, List<OntProperty> properties, Individual rangeIndividual) {
        Iterator<OntProperty> it = properties.iterator();
        OntProperty property = null;
        while (it.hasNext()) {
            property = it.next();
            //System.err.println(domainIndividual.toString());
            //System.err.println(property.toString());
            if (this.existStatement(domainIndividual, property, rangeIndividual)) {
                return true;
            }
            if (this.existStatement(rangeIndividual, this.getInverse(property), domainIndividual)) {
                return true;
            }
        }
        return false;
    }

    /**
     * This method is used to see if a statement has already been added
     * to the model, between the property given. This is used by methods
     * to check whether functional properties are not violated, even in a hidden form
     * 
     * @param domainIndividual
     * @param property
     * @param rangeIndividual
     * @return
     */
    private boolean existStatement(Individual domainIndividual, OntProperty property, Individual rangeIndividual) {
        if (property == null) {
            return false;
        }

        StmtIterator sit = model.listStatements(domainIndividual, null, rangeIndividual);
        Statement stmt = null;
        List<Property> existingProperties = new ArrayList<Property>();
        while (sit.hasNext()) {
            stmt = sit.nextStatement();
            existingProperties.add(stmt.getPredicate());
        }

        //List superProperties = this.getSuperProperties(property);
        //List subProperties = this.getSubProperties(property);
        List<OntProperty> properties = this.getPropertyTree(property);

        Iterator<Property> it = existingProperties.iterator();
        Property pp = null;
        while (it.hasNext()) {
            pp = it.next();

            /*
            Iterator it1 = superProperties.iterator();
            while (it1.hasNext()) {
            Property p1 = (Property) it1.next();
            if (p1.getURI().equalsIgnoreCase(pp.getURI())) {
            return true;
            }
            }
            Iterator it2 = subProperties.iterator();
            while (it2.hasNext()) {
            Property p2 = (Property) it2.next();
            if (p2.getURI().equalsIgnoreCase(pp.getURI())) {
            return true;
            }
            }
             * */
            Iterator<OntProperty> it1 = properties.iterator();
            Property p1 = null;
            while (it1.hasNext()) {
                p1 = (OntProperty) it1.next();
                if (p1.getURI().equalsIgnoreCase(pp.getURI())) {
                    return true;
                }
            }
            if (property.getURI().equalsIgnoreCase(pp.getURI())) {
                return true;
            }
        }
        return false;
    }

    /**
     * Method for actually linking and inserting into the model the
     * relationships between the parameters provided. Takes into account the
     * Functional, Inverse Functional and Symmetric properties where needed
     *
     * @param property
     * @param domainIndividual
     * @param rangeIndividual
     * @param domain
     * @param range
     */
    private void doLinking(OntProperty property, Individual domainIndividual, Individual rangeIndividual, OntClass domain, OntClass range) {
        if (true && !this.isFunctionalProperty(property) && !this.isInverseFunctionalProperty(property)) {
            domainIndividual.addProperty(property, rangeIndividual);
            feedBack.addRelationInstance(domain, range, property);
        } else if (true) {
            List<OntProperty> propertyTree = this.getPropertyTree(property);
            List<OntProperty> inversePropertyTree = this.getPropertyTree(this.getInverse(property));
            if (!this.isInverseFunctionalProperty(property)) {
                if (this.checkFucntionalInstantiatedProperty(propertyTree, domainIndividual)
                        && this.checkFucntionalInstantiatedProperty(inversePropertyTree, rangeIndividual)) {
                    domainIndividual.addProperty(property, rangeIndividual);
                    feedBack.addRelationInstance(domain, range, property);
                    this.setFunctionalInstantiatedProperty(propertyTree, domainIndividual);
                    this.setFunctionalInstantiatedProperty(inversePropertyTree, rangeIndividual);
                }
            } else {
                if (this.checkFucntionalInstantiatedProperty(propertyTree, domainIndividual)
                        && this.checkFucntionalInstantiatedProperty(inversePropertyTree, rangeIndividual)
                        && this.checkFucntionalInstantiatedProperty(propertyTree, rangeIndividual)
                        && this.checkFucntionalInstantiatedProperty(inversePropertyTree, domainIndividual)) {
                    domainIndividual.addProperty(property, rangeIndividual);
                    feedBack.addRelationInstance(domain, range, property);
                    this.setFunctionalInstantiatedProperty(propertyTree, domainIndividual);
                    this.setFunctionalInstantiatedProperty(inversePropertyTree, rangeIndividual);
                    this.setFunctionalInstantiatedProperty(propertyTree, rangeIndividual);
                    this.setFunctionalInstantiatedProperty(inversePropertyTree, domainIndividual);
                }
            }
        }
    }

    /**
     * This method checks a given set of properties and an individual, whether
     * one of the properties in the list was already used previously with this
     * Individual
     *
     * @param properties
     * @param individual
     * @return
     */
    private boolean checkFucntionalInstantiatedProperty(List<OntProperty> properties, Individual individual) {
        Iterator<OntProperty> it = properties.iterator();
        Property p = null;
        int pindex = -1;
        int iindex = Integer.parseInt(individual.getURI().substring(individual.getURI().indexOf("#Individual") + 11));
        while (it.hasNext()) {
            p = it.next();
            pindex = Integer.parseInt(p.getURI().substring(p.getURI().indexOf("#ObjectProperty") + 15));
            if (this.instantiatedFunctionals[iindex][pindex]) {
                return false;
            }
        }
        return true;
    }

    /**
     * This method sets the correct cells in the table to true for the given
     * property-list and individual
     * 
     * @param properties
     * @param individual
     */
    private void setFunctionalInstantiatedProperty(List<OntProperty> properties, Individual individual) {
        Iterator<OntProperty> it = properties.iterator();
        Property p = null;
        int pindex = -1;
        int iindex = Integer.parseInt(individual.getURI().substring(individual.getURI().indexOf("#Individual") + 11));
        while (it.hasNext()) {
            p = it.next();
            pindex = Integer.parseInt(p.getURI().substring(p.getURI().indexOf("#ObjectProperty") + 15));
            this.instantiatedFunctionals[iindex][pindex] = true;
        }
    }

    /**
     * Method for actually linking and inserting into the model the
     * relationships between the parameters provided. Takes into account the
     * Functional, Inverse Functional and Symmetric properties where needed
     * 
     * @param property
     * @param domainIndividual
     * @param rangeIndividual
     * @param domain
     * @param range
     */
    /*
    private void doLinking(OntProperty property, Individual domainIndividual, Individual rangeIndividual, OntClass domain, OntClass range) {

    if (true && !this.isFunctionalProperty(property) && !this.isInverseFunctionalProperty(property)) {
    domainIndividual.addProperty(property, rangeIndividual);
    feedBack.addRelationInstance(domain, range, property);
    } else if (true && this.isFunctionalProperty(property) && !this.isInverseFunctionalProperty(property) && !this.isSymmetricProperty(property)) {
    Individual nully = null;
    // With addition for complete tree traversel
    if (!this.existStatement(domainIndividual, property, nully) && !this.existStatement(null, this.getInverse(property), domainIndividual) && !this.existStatement(domainIndividual, this.getPropertyTree(property), nully)) {
    try {
    if (this.getInverse(property) != null) {
    if (!this.existStatement(null, this.getInverse(property), rangeIndividual)) {
    domainIndividual.setPropertyValue(property, rangeIndividual);
    feedBack.addRelationInstance(domain, range, property);
    }
    } else {
    domainIndividual.setPropertyValue(property, rangeIndividual);
    feedBack.addRelationInstance(domain, range, property);
    }
    } catch (Exception e) {
    domainIndividual.setPropertyValue(property, rangeIndividual);
    feedBack.addRelationInstance(domain, range, property);
    }
    }
    } else if (true && !this.isFunctionalProperty(property) && this.isInverseFunctionalProperty(property) && !this.isSymmetricProperty(property)) {
    Individual nully = null;
    if ((property.getURI().endsWith("86") || property.getURI().endsWith("83")) && rangeIndividual.getURI().endsWith("753")){
    System.out.println(domainIndividual);
    System.out.println(property);
    }
    if (!this.existStatement(null, property, rangeIndividual) && !this.existStatement(rangeIndividual, this.getInverse(property), nully)) {
    try {
    //if (property.getInverse() != null) {
    if (this.getInverse(property) !=null) {
    //if (!this.existStatement(domainIndividual, property.getInverse(), nully)) {
    if (!this.existStatement(rangeIndividual, this.getInverse(property), nully)) {
    domainIndividual.setPropertyValue(property, rangeIndividual);
    feedBack.addRelationInstance(domain, range, property);
    System.out.println("1");
    } else {
    System.out.println("NOT 1");
    }
    } else {
    domainIndividual.setPropertyValue(property, rangeIndividual);
    feedBack.addRelationInstance(domain, range, property);
    System.out.println("2");
    }

    } catch (Exception e) {
    domainIndividual.setPropertyValue(property, rangeIndividual);
    feedBack.addRelationInstance(domain, range, property);
    System.out.println("3");
    }

    /**
    InfModel inf = ModelFactory.createOntologyModel(PelletReasonerFactory.THE_SPEC, model);
    ValidityReport report = inf.validate();
     * */
    //System.out.println(domainIndividual + " + " + property + " + " + rangeIndividual);
                /*
    System.out.println(report.isValid());

    }
    } else if (false && this.isFunctionalProperty(property) && this.isInverseFunctionalProperty(property) && !this.isSymmetricProperty(property)) {
    Individual nully = null;
    boolean ok = true;
    if (!this.existStatement(domainIndividual, property, nully)) {
    if (!this.existStatement(null, property, rangeIndividual)) {

    if (this.isSymmetricProperty(property)) {
    if (this.existStatement(rangeIndividual, property, nully)) {
    ok = false;
    }
    if (!this.existStatement(null, property, domainIndividual)) {
    ok = false;
    }
    }

    if (ok) {
    try {
    if (property.getInverse() != null) {
    if (!this.existStatement(null, property, rangeIndividual)) {
    try {
    if (property.getInverse() != null) {
    if (!this.existStatement(domainIndividual, property.getInverse(), nully)) {
    domainIndividual.setPropertyValue(property, rangeIndividual);
    feedBack.addRelationInstance(domain, range, property);
    }
    } else {
    domainIndividual.setPropertyValue(property, rangeIndividual);
    feedBack.addRelationInstance(domain, range, property);
    }
    } catch (Exception e) {
    }
    }
    } else {
    domainIndividual.setPropertyValue(property, rangeIndividual);
    feedBack.addRelationInstance(domain, range, property);
    }
    } catch (Exception e) {
    }
    }
    }
    }
    }
    }
     */
    /**
     * Takes a random individual belonging to the domnain and the range of
     * the supplied property and links it together.
     * 
     * @param property
     * @param domain
     * @param range
     */
    private void linkInterCluster(OntProperty property, OntClass domain,
            OntClass range) {
        OntoList ontoList = new OntoList();
        ontoList.addAll(domain.listInstances().toList());
        if (ontoList.size() != 0) {
            Individual domainIndividual = ontoList.getIndividual(pseudoRand.nextInt(ontoList.size()));
            OntoList ontoList2 = new OntoList();
            ontoList2.addAll(range.listInstances().toList());
            if (ontoList2.size() != 0) {
                Individual rangeIndividual = ontoList2.getIndividual(pseudoRand.nextInt(ontoList2.size()));
                this.doLinking(property, domainIndividual, rangeIndividual, domain, range);
            }
        }
    }

    /**
     * Gets all properties and keep those linking classes from different
     * clusters
     * 
     * @return Set<OntProperty> interClusterProperties
     */
    private List<OntProperty> getInterClusterProperties() {
        ObjectProperty property = null;
        OntResource domain = null;
        OntResource range = null;
        List<OntProperty> interClusterProperties = new ArrayList<OntProperty>();
        ExtendedIterator it = model.listObjectProperties();
        while (it.hasNext()) {
            property = (ObjectProperty) it.next();
            domain =
                    property.getDomain();
            range =
                    property.getRange();
            if (domain != null && range != null) {
                if (this.isInterClusterRelation(domain, range)) {
                    interClusterProperties.add(property);
                }

            } else {
                DeferredInterClusterOntProperty deferredInterClusterOntProperty = new DeferredInterClusterOntProperty();
                deferredInterClusterOntProperty.setOntProperty(property);
                this.deferredInterclusterProperties.add(deferredInterClusterOntProperty);
            }

        }
        return interClusterProperties;
    }

    /**
     * Checks whether the domain and the range classes (or its members) reside
     * in different clusters?
     * 
     * @param domain
     * @param range
     * @return boolean
     */
    private boolean isInterClusterRelation(OntResource domain, OntResource range) {
        if (domain.isURIResource()) {
            return this.isInterClusterRelation(model.getOntClass(domain.getURI()), range);
        } else {
            boolean interCluster = true;
            // The order in which the ontClasses are checked whether they
            // belong to different clusters doesn't affect to final conclusion
            // no change needed !
            if (domain.asClass().isEnumeratedClass()) {
                EnumeratedClass eClass = domain.asClass().asEnumeratedClass();
                ExtendedIterator eit = eClass.listOneOf();
                while (eit.hasNext() && interCluster) {
                    interCluster = this.isInterClusterRelation((OntClass) eit.next(), range);
                }

            } else if (domain.asClass().isUnionClass()) {
                UnionClass uClass = domain.asClass().asUnionClass();
                ExtendedIterator eit = uClass.listOperands();
                while (eit.hasNext() && interCluster) {
                    interCluster = this.isInterClusterRelation((OntClass) eit.next(), range);
                }

            }
            return interCluster;
        }

    }

    /**
     * Checks whether the domain and the range classes (or its members) reside
     * in different clusters?
     * 
     * @param domain
     * @param range
     * @return boolean
     */
    private boolean isInterClusterRelation(OntClass domain, OntResource range) {
        if (range.isURIResource()) {
            return this.isInterClusterRelation(domain, model.getOntClass(range.getURI()));
        } else {
            // The order in which the ontClasses are checked whether they
            // belong to different clusters doesn't affect to final conclusion
            // no change needed !            
            boolean interCluster = true;
            if (range.asClass().isEnumeratedClass()) {
                EnumeratedClass eClass = range.asClass().asEnumeratedClass();
                ExtendedIterator eit = eClass.listOneOf();
                while (eit.hasNext() && interCluster) {
                    interCluster = this.isInterClusterRelation(domain,
                            (OntClass) eit.next());
                }

            } else if (range.asClass().isUnionClass()) {
                UnionClass uClass = range.asClass().asUnionClass();
                ExtendedIterator eit = uClass.listOperands();
                while (eit.hasNext() && interCluster) {
                    interCluster = this.isInterClusterRelation(domain,
                            (OntClass) eit.next());
                }

            }
            return interCluster;
        }

    }

    /**
     * Checks whether the domain and the range classes (or its members) reside
     * in different clusters?
     * 
     * @param domain
     * @param range
     * @return boolean
     */
    private boolean isInterClusterRelation(OntClass domain, OntClass range) {
        List<Cluster> clusters = feedBack.getGeneratedClusters();
        Iterator<Cluster> it = clusters.iterator();
        Cluster cluster = null;
        boolean interCluster = true;
        // The order in which the ontClasses are checked whether they
        // belong to different clusters doesn't affect to final conclusion
        // no change needed !
        while (it.hasNext() && interCluster) {
            cluster = it.next();
            if (cluster.getCluster().contains(domain)
                    && cluster.getCluster().contains(range)) {
                interCluster = false;
            }

        }
        return interCluster;
    }

    /**
     * Method for generating datatypeproperties. Again the parameters from
     * the config file are used to generate sufficient instances.
     * 
     * Also functionality is taken into account
     */
    private void generateDatatypePropertyInstances() {

        //Create a list of all NamedClasses to be considered
        ExtendedIterator it = model.listNamedClasses();
        OntoList ontoList = new OntoList();
        ontoList.addAll(it.toList());

        ExtendedIterator pit = null;
        OntoList ontoList2 = new OntoList();
        OntClass ontClass = null;
        OntProperty property = null;
        int datatypePropertyIndividualCount = 0;
        int generatedPropertyIndividualCount = 0;
        int totalPropertyCount = 0;
        int pCount = 0;
        Individual individual = null;
        Literal literal = null;
        int count = 0;

        for (int i = 0; i
                < ontoList.size(); i++) {
            //Define how many individuals of the currently considered class
            //should have their datatype properties filled in
            ontClass = ontoList.getOntClass(i);
            datatypePropertyIndividualCount =
                    ontClass.listInstances().toList().size();
            datatypePropertyIndividualCount *=
                    this.individualDatatypePropertyPercentage;
            datatypePropertyIndividualCount /=
                    100;

            //Get a List of all Datatype Properties (incl. superproperties)
            pit =
                    ontClass.listDeclaredProperties();
            ontoList2 =
                    new OntoList();
            ontoList2.clear();
            ontoList2.addAll(pit.toList());

            Iterator<OntClass> superit = this.getSuperClasses(ontClass).iterator();
            OntClass superOntClass = null;
            while (superit.hasNext()) {
                superOntClass = (OntClass) superit.next();
                ontoList2.addAll(superOntClass.listDeclaredProperties().toList());
            }

            //For each of those properties, do the instantiation procedure
            for (int j = 0; j
                    < ontoList2.size(); j++) {
                property = ontoList2.getOntProperty(j);
                //Only consider datatype properties
                if (property.isDatatypeProperty()) {
                    //as long as not sufficient individuals have their property filled in
                    while (generatedPropertyIndividualCount
                            < datatypePropertyIndividualCount) {
                        //Create a list of all individuals of the class
                        OntoList is = new OntoList();
                        is.addAll(ontClass.listInstances().toList());
                        individual =
                                is.getIndividual(pseudoRand.nextInt(is.size()));
                        if (debug) {
                            System.out.println("[GENERATEPROPERTYINSTANCES][DEBUG] "
                                    + "IndividualCount: " + ontClass.listInstances().toList().size());
                        }

                        while (pCount < this.datatypePropertyCount) {
                            if (debug) {
                                System.out.println("[GENERATEPROPERTYINSTANCES][DEBUG] "
                                        + "Range of DatatypeProperty: " + property.getRange() + " " + pCount);
                            }

                            if (property.getRange().toString().endsWith("String") || property.getRange().toString().endsWith("string")) {
                                literal = model.createTypedLiteral("String" + (totalPropertyCount + generatedPropertyIndividualCount + pCount));
                            } else {
                                literal = model.createTypedLiteral(totalPropertyCount + generatedPropertyIndividualCount + pCount);
                            }

                            if (!this.isFunctionalDatatypeProperty(property)) {
                                individual.addProperty(property, literal);
                                feedBack.addRelationInstance(ontClass, property.getRange(), property);
                                pCount++;
                            } else if (individual.getPropertyValue(property) == null) {
                                if (this.checkAndAttachFunctionalProperty(property, individual, literal, ontClass)) {
                                    pCount = this.datatypePropertyCount;
                                } else {
                                    pCount = this.datatypePropertyCount;
                                }
                            } else {
                                pCount = this.datatypePropertyCount;
                            }

                        }
                        pCount = 0;
                        generatedPropertyIndividualCount++;

                    }







                    totalPropertyCount += generatedPropertyIndividualCount;
                    generatedPropertyIndividualCount =
                            0;
                }

            }
        }
    }

    /**
     * Due to the inability of Jena to also infer the functioanlity
     * characteristic of the parent properties without an attached reasoner
     * we have to do this by ourselves.
     * 
     * @param property
     * @return
     */
    private boolean isFunctionalProperty(OntProperty property) {
        boolean isFunctional = false;
        if (property.isFunctionalProperty()) {
            isFunctional = true;
        }

        ExtendedIterator it = property.listSuperProperties();
        OntProperty temp = null;
        while (it.hasNext()) {
            temp = (OntProperty) it.next();
            if (isFunctionalProperty(temp)) {
                isFunctional = true;
            }

        }

        try {
            if (property != null) {
                if (property.getInverseOf() != null) {
                    //return this.isFunctionalProperty(property.getInverseOf());
                    if (this.isInverseFunctionalProperty(property.getInverseOf())) {
                        isFunctional = true;
                    }

                }
            }
        } catch (Exception e) {
        }
        return isFunctional;
    }

    /**
     * Due to the inability of Jena to also infer the functioanlity
     * characteristic of the parent properties without an attached reasoner
     * we have to do this by ourselves.
     *
     * @param property
     * @return
     */
    private boolean isFunctionalDatatypeProperty(OntProperty property) {
        boolean isFunctional = false;
        if (property.isFunctionalProperty()) {
            isFunctional = true;
        }

        List<OntProperty> propertyTree = this.getPropertyTree(property);
        Iterator<OntProperty> it = propertyTree.iterator();

        //ExtendedIterator it = property.listSuperProperties();
        OntProperty temp = null;
        while (it.hasNext()) {
            temp = (OntProperty) it.next();
            if (temp.isFunctionalProperty()) {
                //isFunctional = true;
                return true;
            }

        }

        try {
            if (property != null) {
                if (property.getInverseOf() != null) {
                    //return this.isFunctionalProperty(property.getInverseOf());
                    if (property.getInverseOf().isFunctionalProperty()) {
                        isFunctional = true;
                    }

                }
            }
        } catch (Exception e) {
        }
        return isFunctional;
    }

    /**
     * Due to the inability of Jena to also infer the symmetric
     * characteristic of the parent properties without an attached reasoner
     * we have to do this by ourselves.
     * @param property
     * @return
     */
    private boolean isSymmetricProperty(OntProperty property) {
        boolean isSymmetric = false;
        if (property.isSymmetricProperty()) {
            isSymmetric = true;
        }

        ExtendedIterator it = property.listSuperProperties();
        OntProperty temp = null;
        while (it.hasNext()) {
            temp = (OntProperty) it.next();
            if (isSymmetricProperty(temp)) {
                isSymmetric = true;
            }

        }

        return isSymmetric;
    }

    /**
     * Due to the inability of Jena to also infer the inverse functioanlity
     * characteristic of the parent properties without an attached reasoner
     * we have to do this by ourselves.
     * @param property
     * @return
     */
    private boolean isInverseFunctionalProperty(OntProperty property) {
        boolean isInverseFunctional = false;
        if (property.isInverseFunctionalProperty()) {
            isInverseFunctional = true;
        }

        ExtendedIterator it = property.listSuperProperties();
        OntProperty temp = null;
        while (it.hasNext()) {
            temp = (OntProperty) it.next();
            if (this.isInverseFunctionalProperty(temp)) {
                isInverseFunctional = true;
            }
        }

        try {
            if (property != null) {
                if (property.getInverseOf() != null) {
                    if (this.isFunctionalProperty(property.getInverseOf())) {
                        isInverseFunctional = true;
                    }
                }

                if (property.getInverse() != null) {
                    if (property.getInverse().isFunctionalProperty()) {
                        isInverseFunctional = true;
                    }
                }

            }
        } catch (Exception e) {
        }
        return isInverseFunctional;
    }

    private OntProperty getInverse(OntProperty property) {
        if (property.getInverse() != null) {
            return property.getInverse();
        } else {
            ExtendedIterator it = model.listObjectProperties();
            OntProperty invprop = null;
            while (it.hasNext()) {
                invprop = (OntProperty) it.next();
                if (invprop.getInverse() != null && invprop.getInverse().getURI().equalsIgnoreCase(property.getURI())) {
                    return invprop;
                }
            }
        }

        //Do the same, but for the super properties!!
        List<OntProperty> superProperties = this.getSuperProperties(property);
        Iterator<OntProperty> superit = superProperties.iterator();
        OntProperty superProperty = null;
        while (superit.hasNext()) {
            superProperty = superit.next();

            if (superProperty.getInverse() != null) {
                return superProperty.getInverse();
            } else {
                ExtendedIterator it = model.listObjectProperties();
                OntProperty invprop = null;
                while (it.hasNext()) {
                    invprop = (OntProperty) it.next();
                    if (invprop.getInverse() != null && invprop.getInverse().getURI().equalsIgnoreCase(superProperty.getURI())) {
                        return invprop;
                    }
                }
            }
        }

        return null;
    }

    /**
     * This method is called in a second iteration, after the domain and
     * ranges of properties attached to enumerated classes have been set.
     */
    public void generateEnumRelations() {
        //Cluster-classes
        //As the implementation of this list is an ArrayList, the order is
        //deterministic. No other sorting needed!
        Iterator<DeferredClusterOntClass> it = deferredClusterClasses.iterator();
        DeferredClusterOntClass ontClass = null;
        while (it.hasNext()) {
            ontClass = it.next();
            //System.out.println("Doing OntClass: " + ontClass);
            if (ontClass.getOntClass().isURIResource()) {
                this.linkOntClass(ontClass.getOntClass(), ontClass.getIndividualConnectionCount(),
                        ontClass.getCluster());
            }
        }

        //InterCluster-properties
        //As the implementation of this list is an ArrayList, the order is
        //deterministic. No other sorting needed!
        Iterator<DeferredInterClusterOntProperty> it2 = deferredInterclusterProperties.iterator();
        DeferredInterClusterOntProperty ontProperty = null;
        while (it2.hasNext()) {
            ontProperty = it2.next();
            //System.out.println(ontProperty);
            int c = 0;
            while (c < ontProperty.getIndividualConnectionCount()) {
                //System.out.println("************************");
                //System.out.println(ontProperty.getOntProperty());
                //System.out.println(ontProperty.getOntProperty().getDomain());
                //System.out.println(ontProperty.getOntProperty().getRange());
                if (ontProperty.getOntProperty().getDomain().isURIResource()
                        && ontProperty.getOntProperty().getRange().isURIResource()) {
                    this.linkInterCluster(ontProperty.getOntProperty(),
                            ontProperty.getOntProperty().getDomain(),
                            ontProperty.getOntProperty().getRange());
                }
                c++;
            }

            if (ontProperty.getOntProperty().getDomain().isURIResource()
                    && ontProperty.getOntProperty().getRange().isURIResource()) {
                this.linkInterClusterSubClasses(ontProperty.getOntProperty(),
                        ontProperty.getOntProperty().getDomain(),
                        ontProperty.getOntProperty().getRange(),
                        ontProperty.getIndividualConnectionCount());
            }
            c = 0;
        }
    }

    /**
     * Get some feedback about the generated A-Box (data level) of the ontology.
     * E.g. the number of instances created, the clusters and so on.
     * @see AboxInformation
     * @return ABoxInformation: an object containing all the information about the
     *                          generated A-Box.
     */
    public AboxInformation getFeedBack() {
        return feedBack;
    }

    /**
     * Method used to define whether a certain Literal can be added
     * through a DataType property without the functionality constraints 
     * being violated.
     * 
     * @param property
     * @param individual
     * @param literal
     * @param ontClass
     * @return
     */
    private boolean checkAndAttachFunctionalProperty(OntProperty property, Individual individual, Literal literal, OntClass ontClass) {
        boolean canAdd = true;

        /**
        List<OntProperty> superProperties = this.getSuperProperties(property);
        //ExtendedIterator prit = property.listSuperProperties();
        Iterator prit = superProperties.iterator();
        OntProperty p = null;
        while (prit.hasNext()) {
        p = (OntProperty) prit.next();
        if (individual.getPropertyValue(p) != null) {
        canAdd = false;
        }

        }
        List<OntProperty> subProperties = this.getSubProperties(property);
        //prit = property.listSubProperties();
        prit = subProperties.iterator();
        while (prit.hasNext()) {
        p = (OntProperty) prit.next();
        if (individual.getPropertyValue(p) != null) {
        canAdd = false;
        }

        }
         * */
        List<OntProperty> properties = this.getPropertyTree(property);
        Iterator<OntProperty> prit = properties.iterator();
        OntProperty p = null;
        while (prit.hasNext()) {
            p = (OntProperty) prit.next();
            if (individual.getProperty(p) != null) {
                canAdd = false;
            }
        }

        /**
        if (canAdd) {
        canAdd = this.canAddSameLevelProperties(individual, property);
        }
         * */
        if (canAdd) {
            individual.setPropertyValue(property, literal);
            feedBack.addRelationInstance(ontClass, property.getRange(), property);
        }

        return canAdd;
    }

    /**
     * See whether the superproperties already have a Literal defined for 
     * a given functioanl property
     * 
     * @param individual
     * @param property
     * @return
     */
    private boolean canAddSuperProperties(Individual individual, OntProperty property) {
        ExtendedIterator prit = property.listSuperProperties();
        OntProperty p = null;
        boolean canAdd = true;
        while (prit.hasNext()) {
            p = (OntProperty) prit.next();
            canAdd =
                    this.canAddSuperProperties(individual, p);
            if (canAdd && individual.getPropertyValue(p) != null) {
                canAdd = false;
            }

        }
        if (property.listSuperProperties().toList().size() == 0) {
            return true;
        }

        return canAdd;
    }

    /**
     * Checks whether the subproperties already have a Literal set for a given
     * datatypeproperty, so that the functioanlity is not violateed
     * 
     * @param individual
     * @param property
     * @return
     */
    boolean canAddSubProperties(Individual individual, OntProperty property) {
        ExtendedIterator prit = property.listSubProperties();
        OntProperty p = null;
        boolean canAdd = true;
        while (prit.hasNext()) {
            p = (OntProperty) prit.next();
            canAdd =
                    this.canAddSubProperties(individual, p);
            if (canAdd && individual.getPropertyValue(p) != null) {
                canAdd = false;
            }

        }
        if (property.listSubProperties().toList().size() == 0) {
            return true;
        }

        return canAdd;
    }

    /**
     * Also the same level properties have to be checked for functioanilty
     * constraints
     * 
     * @param individual
     * @param property
     * @return
     */
    private boolean canAddSameLevelProperties(Individual individual, OntProperty property) {
        boolean canAdd = true;
        ExtendedIterator it = property.listSuperProperties();
        OntProperty p = null;
        ExtendedIterator iit = null;
        OntProperty pp = null;
        while (it.hasNext()) {
            p = (OntProperty) it.next();
            iit =
                    p.listSubProperties();
            while (iit.hasNext()) {
                pp = (OntProperty) iit.next();
                if (individual.getPropertyValue(pp) != null) {
                    canAdd = false;
                }

            }
        }
        return canAdd;
    }

    /**
     * Due to the inability of Jena to also infer the SubProperties
     * of the parent properties without an attached reasoner
     * we have to do this by ourselves.
     * @param property
     * @return
     */
    private List<OntProperty> getSubProperties(OntProperty property) {
        List<OntProperty> newList = new ArrayList<OntProperty>();
        ExtendedIterator eit = property.listSubProperties();
        while (eit.hasNext()) {
            OntProperty eitp = (OntProperty) eit.next();
            newList.addAll(this.getSubProperties(eitp));
        }

        newList.add(property);
        return newList;
    }

    /**
     * Due to the inability of Jena to also infer the SuperProperties
     * of the parent properties without an attached reasoner
     * we have to do this by ourselves.
     * @param property
     * @return
     */
    private List<OntProperty> getSuperProperties(OntProperty property) {
        List<OntProperty> newList = new ArrayList<OntProperty>();
        ExtendedIterator eit = property.listSuperProperties();
        while (eit.hasNext()) {
            OntProperty eitp = (OntProperty) eit.next();
            newList.addAll(this.getSuperProperties(eitp));
        }

        newList.add(property);
        return newList;
    }

    /**
     * Due to the inability of Jena to also infer the PropertyTree
     * of the parent properties without an attached reasoner
     * we have to do this by ourselves.
     * @param property
     * @return
     */
    private List<OntProperty> getPropertyTree(OntProperty property) {
        List<OntProperty> newList = new ArrayList<OntProperty>();
        OntProperty superProperty = null;
        OntProperty temp = null;
        temp = property;
        if (property == null) {
            return newList;
        }
        superProperty = property.getSuperProperty();
        while (superProperty != null) {
            temp = superProperty;
            superProperty = superProperty.getSuperProperty();
        }
        newList.addAll(this.getPropertySubTree(temp));
        return newList;
    }

    /**
     * Due to the inability of Jena to also infer the PropertyTree
     * of the parent properties without an attached reasoner
     * we have to do this by ourselves. (Recursive help-method)
     * @param property
     * @return
     */
    private List<OntProperty> getPropertySubTree(OntProperty property) {
        List<OntProperty> subProperties = new ArrayList<OntProperty>();
        if (property.listSubProperties().toList().size() == 0) {
            subProperties.add(property);
            return subProperties;
        } else {
            ExtendedIterator it = property.listSubProperties();
            subProperties.add(property);
            while (it.hasNext()) {
                subProperties.addAll(this.getPropertySubTree((OntProperty) it.next()));
            }
        }
        return subProperties;
    }

    /**
     * Due to the inability of Jena to also infer the SubClasses
     * of the parent properties without an attached reasoner
     * we have to do this by ourselves.
     * @param ontClass
     * @return
     */
    private List<OntClass> getSubClasses(OntClass ontClass) {
        List<OntClass> newList = new ArrayList<OntClass>();
        ExtendedIterator eit = ontClass.listSubClasses();
        while (eit.hasNext()) {
            OntClass eitc = (OntClass) eit.next();
            newList.addAll(this.getSubClasses(eitc));
        }

        newList.add(ontClass);
        return newList;
    }

    /**
     * Due to the inability of Jena to also infer the SuperClasses
     * of the parent properties without an attached reasoner
     * we have to do this by ourselves.
     * @param ontClass
     * @return
     */
    private List<OntClass> getSuperClasses(OntClass ontClass) {
        List<OntClass> newList = new ArrayList<OntClass>();
        ExtendedIterator eit = ontClass.listSuperClasses();
        while (eit.hasNext()) {
            OntClass eitc = (OntClass) eit.next();
            newList.addAll(this.getSuperClasses(eitc));
        }

        newList.add(ontClass);
        return newList;
    }
}
