/*
 * Copyright 2008-2011 Femke Ongenae, Stijn Verstichel
 * distributed under the terms of the AGPL License
 * 
 * This file is part of OTAGen.
 * 
 * OTAGen is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Femke Ongenae & Stijn Verstichel
 * Ghent University - INTEC - IBCN
 * Zuiderpoort Block C0 bus 201
 * Gaston Crommenlaan 8
 * B-9050 Ghent
 * Belgium
 * E-mail: Femke.Ongenae@intec.ugent.be
 * E-mail: Stijn.Verstichel@intec.ugent.be
 * 
 */

package otagen;

import com.hp.hpl.jena.ontology.AllValuesFromRestriction;
import com.hp.hpl.jena.ontology.CardinalityRestriction;
import com.hp.hpl.jena.ontology.DatatypeProperty;
import com.hp.hpl.jena.ontology.MaxCardinalityRestriction;
import com.hp.hpl.jena.ontology.MinCardinalityRestriction;
import com.hp.hpl.jena.ontology.ObjectProperty;
import com.hp.hpl.jena.ontology.OntClass;
import com.hp.hpl.jena.ontology.OntModel;
import com.hp.hpl.jena.ontology.OntProperty;
import com.hp.hpl.jena.ontology.SomeValuesFromRestriction;
import com.hp.hpl.jena.rdf.model.RDFList;
import com.hp.hpl.jena.util.iterator.ExtendedIterator;
import java.util.ArrayList;
import java.util.Random;
import otagen.datatypes.OntoList;
import otagen.feedback.ClassInformation;

/**
 * Creates the class hierarchie of the ontology.
 *
 * @author Femke Ongenae.
 * @author Stijn Verstichel.
 */
public class ClassesGen {
    
    /////-------------------------FIELDS-------------------------/////
    
    //The base URI of the ontology
    private static String baseURI = null;

    //the number of classes
    private int nrOfClasses = 0;
    
    //the number of classes which aren't logically defined.
    private int nrOfSimpleClasses = 0;
    
    //the generated ontology
    private OntModel model = null;
    
    //a pseudo random generator (with seed);
    private Random pseudoRand = null;
    
    //the maximal depth of the classes and properties
    private int maxDepth = 0;
    
    //true : create logical defined classes,
    //false : create some empty (non-defined) classes.
    private boolean flag = false;
    
    //the number of classes of the previous category that must be created
    private int nrOfFlagClasses = 0;
    
    //the maximal number of subclasses a class can have.
    private int maxNrSubClasses = 0;
    
    //holds the information about the generated classes.
    private ClassInformation classInformation = null;
    
    //holds the number of enumerated classes that have been created.
    private int nrOfEnumClasses = 0;
    
    //holds the number of restriction classes that have been created.
    private int nrOfRestrictionClasses = 0;
    
    //holds the object properties used to create the restriction classes.
    private ArrayList<OntProperty> restrictionProps = new ArrayList<OntProperty>();
    
    //holds the classes for which properties should be generated.
    private ArrayList<OntClass> connectingClasses = new ArrayList<OntClass>();
    
    /////-------------------------Constructor-------------------------/////
    
    /** 
     * Creates the class structure of the ontology (classes + hierarchie)
     *
     * @param OntModel model: the ontology we're creating.
     * @param String baseURI: the base URI of the ontology.
     * @param int nrOfClasses: the number of classes.
     * @param Random pseudoRand: a pseudo random generator (with seed).
     * @param int maxDepth: the maximal depth of the classes and properties.
     * @param boolean flag: A flag: true : create logical defined classes,
     *                              false : create some empty (non-defined) classes.
     * @param int nrOfFlagClasses: the number of classes of the previous category that must be created.
     * @param int maxNrSubClasses: the maximal number of subclasses a class can have.
     */
    public ClassesGen(OntModel model, String baseURI, int nrOfClasses, Random pseudoRand, 
                      int maxDepth, boolean flag, int nrOfFlagClasses, int maxNrSubClasses) {
        //initialize fields
        this.model = model;
        this.baseURI = baseURI;
        this.nrOfClasses = nrOfClasses;
        this.pseudoRand = pseudoRand;
        this.maxDepth = maxDepth;
        this.flag = flag;
        this.nrOfFlagClasses = nrOfFlagClasses;
        this.maxNrSubClasses = maxNrSubClasses;
        this.classInformation = new ClassInformation();
        
        //number of classes which aren't logically defined
        nrOfSimpleClasses = nrOfClasses - nrOfFlagClasses;
        
        //create the classes that aren't logically defined
        System.out.println("[STATUS] Generating Simple (Not Logical) Classes");
        createSimpleClasses();
        
        //create the logically defined classes
        createLogicalClasses();
    }
    
    /////-------------------------Business Methods-------------------------/////
    
    /**
    * Generates the not logically defined classes in the ontology
    */
    private void createSimpleClasses(){
        //generate the classes/subclassstructure
        int classCounter = 0;
        
        //ArrayList<OntClass> disjointClasses = new ArrayList<OntClass>();
        ArrayList<OntClass> disjointClasses = null;
        while(classCounter < nrOfSimpleClasses){
           int depth = 0;
           //The classes that are al subclasses of owl:Thing have to be disjoint with each other
           classCounter = createSimpleClass(classCounter, depth, null, disjointClasses);
        }
        //Make the classes disjoint with each other
        if(disjointClasses != null){
            for(int i = 0; i < disjointClasses.size(); i++){
                OntClass class1 = disjointClasses.get(i);
                for(int j = i + 1; j < disjointClasses.size(); j++){
                    OntClass class2 = disjointClasses.get(j);
                    class1.addDisjointWith(class2);
                }
            }
        }
    }
    
    /**
     * Method that is recursively used to generate classes and their subclasses.
     *
     * @param int classCounter: counter that counts the number of classes already generated.
     * @param int depth: counter that counts at which depth (in the subclass tree) this class is generated.
     * @param OntClass superClass: -> null : this class doesn't have a superclass (is at the highest level).
     *                             -> != null : the superclass of the class we're currently creating.
     * @param ArrayList<OntClass> disjoint: the group of classes with which the currently created class 
     *                                      should be disjoint. All classes at the highest level are made disjoint.
     * @return int: the number of simple class already created.
     */ 
    private int createSimpleClass(int classCounter, int depth, OntClass superClass, ArrayList<OntClass> disjoint)
    {
        if(classCounter < nrOfSimpleClasses)
        {
            if(classInformation.getSimpleDepth() < depth)
                classInformation.setSimpleDepth(depth);
            //the name of the class (URI) = class + classCounter so we get class0, class1, ...
            OntClass created = model.createClass(baseURI + "class" + classCounter);
            classCounter++;
            
            if(disjoint != null)
                //make the class disjoint of all classes at the same depth in the subtree
                //and with the same superclass
                disjoint.add(created);
            
            if(superClass != null)
            {
                superClass.addSubClass(created);
            }
           
            //Does this class have subclasses?
            int hasSubclass = pseudoRand.nextInt(2);
            //1 = has subclasses, 0 = no subclasses
            //we can't exceed the maximal allowed depth of the ontology
            if(hasSubclass == 1 && depth != maxDepth)
            {
                //how many subclasses does this class try to have?
                int nrOfSubClasses = pseudoRand.nextInt(maxNrSubClasses);
                depth++;
                for(int i = 0; i < nrOfSubClasses; i++)
                {
                    classCounter = createSimpleClass(classCounter, depth, created, null);
                }
            }
            connectingClasses.add(created);
        }
        return classCounter;
    }
    
    /**
     * Generates the logically defined classes in the ontology
     */
    private void createLogicalClasses(){
        //Create some Empty classes (classes with no definition)
        System.out.println("[STATUS] Generating Empty Logical Classes");
        if(!flag)
        {
            int nrOfEmptyClasses = pseudoRand.nextInt(nrOfFlagClasses);
            classInformation.setNrOfEmptyClasses(nrOfEmptyClasses);
            for(int i = 0; i < nrOfEmptyClasses; i++){
                OntClass empty = model.createClass(baseURI + "emptyClass" + i);
                connectingClasses.add(empty);
            }
            //How much logically defined classes do we still have to make?
            nrOfFlagClasses = nrOfFlagClasses - nrOfEmptyClasses;
        }
            
        //logically defined classes are defined based upon the simple classes 
        //create logically defined classes
        System.out.println("[STATUS] Generating Logically Defined Classes");
        //Counters used to give feedback about the generation of the logical classes.
        int nrOfUnionClasses = 0;
        int nrOfIntersectionClasses = 0;
        int nrOfComplementClasses = 0;
        
        //Check if it's possible to create intersection definitions, for
        //that we need two not disjoint classes, so we need a subtree
        //with at least two members
        boolean intersectionPossible = false;
        for(int j = 0; j < nrOfSimpleClasses && !intersectionPossible; j++){
            OntClass ontClass = model.getOntClass(baseURI + "class" + j);
            while(ontClass.hasSuperClass())
                ontClass = ontClass.getSuperClass();
            int count = 0;
            if(ontClass.hasSubClass()){
                ExtendedIterator eit = ontClass.listSubClasses();
                count += eit.toList().size();
                while(eit.hasNext() && count < 2){
                    ontClass = (OntClass) eit.next();
                    if(ontClass.hasSubClass())
                        count ++;
                }
            }
            if(count >= 2)
                intersectionPossible = true;
        }
        
        for(int i = 0; i < nrOfFlagClasses; i++){
            
            int kind;
            if(intersectionPossible)
                //which kind of logically defined class do we want to create?
                kind = pseudoRand.nextInt(5);
            else
                kind = pseudoRand.nextInt(4);
            
            //0 : unionClass
            if(kind == 0){
                //choose 2 random simple classes en generate a new node as the union
                int union1 = pseudoRand.nextInt(nrOfSimpleClasses);
                int union2 = pseudoRand.nextInt(nrOfSimpleClasses);
                OntClass unionClass1 = model.getOntClass(baseURI + "class" + union1);
                OntClass unionClass2 = model.getOntClass(baseURI + "class" + union2);
                RDFList union = model.createList();
                union = union.cons(unionClass1);
                union = union.cons(unionClass2);
                OntClass unionClass = model.createUnionClass(baseURI + "logicalClass" + i,union);
                nrOfUnionClasses++;
                connectingClasses.add(unionClass);
            }
            
            //1 : complementClass
            else if(kind == 1){
                //choose a random class and define this class as the complement
                int compl = pseudoRand.nextInt(nrOfSimpleClasses);
                OntClass complClass = model.getOntClass(baseURI + "class" + compl);
                model.createComplementClass(baseURI + "logicalClass" + i,complClass);
                nrOfComplementClasses++;
            }
           
            //2: restrictionClass
            else if(kind == 2){
                nrOfRestrictionClasses++;
            }
            
            //3: enumeratedClass
            else if(kind == 3){
                nrOfEnumClasses++;
                //createEnumeratedClass(java.lang.String uri, RDFList members)
            }
            
            //4 : intersectionClass
            else if(kind == 4){
                //choose 2 random simple classes en generate a new node as the intersection
                int intersec1 = pseudoRand.nextInt(nrOfSimpleClasses);
                OntClass intersecClass1 = model.getOntClass(baseURI + "class" + intersec1);
                
                //the two classes can't be disjoint (because the intersection would always be
                //empty in that case).
                boolean isDisjoint = true;
                OntClass intersecClass2 = null; 
                while(isDisjoint) {
                    int intersec2 = pseudoRand.nextInt(nrOfSimpleClasses);
                    intersecClass2 = model.getOntClass(baseURI + "class" + intersec2);
                    
                    OntClass superClass = intersecClass2;
                    isDisjoint = false;
                    while(superClass != null && !isDisjoint){
                        isDisjoint = superClass.isDisjointWith(intersecClass1);
                        if(!isDisjoint)
                            isDisjoint = intersecClass1.isDisjointWith(superClass);
                        superClass = superClass.getSuperClass();
                    }
                }
                
                RDFList intersection = model.createList();
                intersection = intersection.cons(intersecClass1);
                intersection = intersection.cons(intersecClass2);
                model.createIntersectionClass(baseURI + "logicalClass" + i,intersection);
                nrOfIntersectionClasses++;
            }
        }
        
        //Configure feedback about the logical classes
        classInformation.setNrOfComplementClasses(nrOfComplementClasses);
        classInformation.setNrOfEnumClasses(nrOfEnumClasses);
        classInformation.setNrOfIntersectionClasses(nrOfIntersectionClasses);
        classInformation.setNrOfUnionClasses(nrOfUnionClasses);
        classInformation.setNrOfRestrictionClasses(nrOfRestrictionClasses);
    }
    
    /**
     * Generates the restrictionclasses: -> cardinality restrictions
     *                                   -> max cardinality restrictions
     *                                   -> min cardinality restrictions
     *                                   -> all values from restrictions
     */
    public void generateRestrictionClasses(){
        //Generate all the properties
        ExtendedIterator it = model.listObjectProperties();
        OntoList properties = new OntoList();
        properties.addAll(it.toList());
        
        it = model.listDatatypeProperties();
        properties.addAll(it.toList());

        //Search all the properties that aren't transitive
        ArrayList<OntProperty> notTransProps = new ArrayList<OntProperty>();
        for(int i=0; i < properties.size(); i++){
            OntProperty prop = properties.getOntProperty(i);
            OntProperty superProp = prop;
            boolean transitive = false;
            while(superProp != null && !transitive){
                transitive = superProp.isTransitiveProperty();
                superProp = superProp.getSuperProperty();
            }
            if(!transitive)
                notTransProps.add(prop);
        }
        //From this search all the properties that aren't functional
        ArrayList<OntProperty> notFuncNotTransProps = new ArrayList<OntProperty>();
        for(int i=0; i < notTransProps.size(); i++){
            OntProperty prop = notTransProps.get(i);
            OntProperty superProp = prop;
            boolean functional = false;
            while(superProp != null && !functional){
                functional = superProp.isFunctionalProperty();
                superProp = superProp.getSuperProperty();
            }
            if(!functional)
                notFuncNotTransProps.add(prop);
        }
        //If no such non functional non transitive properties exist,
        //the cardinality must always be 1
        boolean cardBigger1Possible = true;
        if(notFuncNotTransProps.size() == 0){
            cardBigger1Possible = false;
        }
        
        for(int i=0; i < nrOfRestrictionClasses; i++){
            //choose which restriction will be used
            int kind = pseudoRand.nextInt(5);
            String uri = baseURI + "restrictionClass" + i;
            
            //Generate random cardinality, can be 1,2,3 or 4
            //Unless there are no non functional non transitive properties,
            //in that case the cardinality can only be 1
            int cardinality = 0;
            if(cardBigger1Possible)
                cardinality = pseudoRand.nextInt(4) + 1;
            else
                cardinality = 1;
            
            //Choose random property that is not transitive
            //if kind is 0,1 or 2 and cardinality > 1 it can also not be functional
            //Functional properties can only have cardinality one, so other cardinalities 
            //must be combined with a non functional properties. 
            int prop = 0;
            OntProperty property = null;
             
            if ((kind == 0 || kind == 1 || kind == 2) && cardinality > 1) {
                prop = pseudoRand.nextInt(notFuncNotTransProps.size());
                property = notFuncNotTransProps.get(prop);
            } else {
                prop = pseudoRand.nextInt(notTransProps.size());
                property = notTransProps.get(prop);
            }
            
            restrictionProps.add((OntProperty) property);
            ObjectProperty objProp = null;
            DatatypeProperty datProp = null;
       
            if(property.isObjectProperty())
                objProp = (ObjectProperty) property;
            else
                datProp = (DatatypeProperty) property;                
            
            if(kind == 0){
                //kind = pseudoRand.nextInt(2);
                CardinalityRestriction restriction = null;
                if(objProp != null)
                    restriction = model.createCardinalityRestriction(null,objProp,cardinality);
                else
                    restriction = model.createCardinalityRestriction(null,datProp,cardinality);
                OntClass restrictionClass = model.createClass(uri);
                restriction.addEquivalentClass(restrictionClass);
            }
            if(kind == 1){
                //kind = pseudoRand.nextInt(2);
                MaxCardinalityRestriction restriction = null;
                if(objProp != null)
                    restriction = model.createMaxCardinalityRestriction(null,objProp,cardinality);
                else
                    restriction = model.createMaxCardinalityRestriction(null,datProp,cardinality);
                OntClass restrictionClass = model.createClass(uri);
                restriction.addEquivalentClass(restrictionClass);
            }
            if(kind == 2){
                //kind = pseudoRand.nextInt(2);
                MinCardinalityRestriction restriction = null;
                if(objProp != null)
                   restriction = model.createMinCardinalityRestriction(null,objProp,cardinality);
                else
                   restriction = model.createMinCardinalityRestriction(null,datProp,cardinality);
                OntClass restrictionClass = model.createClass(uri);
                restriction.addEquivalentClass(restrictionClass);
                
            }
            if(kind == 3){
                ExtendedIterator eit = property.listRange();
                OntoList ranges = new OntoList();
                ranges.addAll(eit.toList());
                while(ranges.size() == 0){
                    //Choose random property that is not transitive
                    prop = pseudoRand.nextInt(notTransProps.size());
                    restrictionProps.remove(property);
                    property = notTransProps.get(prop);
                    restrictionProps.add((OntProperty) property);
                    eit = property.listRange();
                    ranges = new OntoList();
                    ranges.addAll(eit.toList());
                }
                int cl = pseudoRand.nextInt(ranges.size());
                OntClass ontClass = ranges.getOntClass(cl);
                AllValuesFromRestriction restriction = null;
                if(objProp != null)
                    restriction = model.createAllValuesFromRestriction(null,objProp,ontClass);
                else
                    restriction = model.createAllValuesFromRestriction(null,datProp,ontClass);
                OntClass restrictionClass = model.createClass(uri);
                restriction.addEquivalentClass(restrictionClass);
            }
            if(kind == 4){
                ExtendedIterator eit = property.listRange();
                OntoList ranges = new OntoList();
                ranges.addAll(eit.toList());
                while(ranges.size() == 0){
                    //Choose random property that is not transitive
                    prop = pseudoRand.nextInt(notTransProps.size());
                    restrictionProps.remove(property);
                    property = notTransProps.get(prop);
                    restrictionProps.add((OntProperty) property);
                    eit = property.listRange();
                    ranges = new OntoList();
                    ranges.addAll(eit.toList());
                }
                int cl = pseudoRand.nextInt(ranges.size());
                OntClass ontClass = ranges.getOntClass(cl);
                SomeValuesFromRestriction restriction = null;
                if(objProp != null)
                    restriction = model.createSomeValuesFromRestriction(null,objProp,ontClass);
                else
                    restriction = model.createSomeValuesFromRestriction(null,datProp,ontClass);
                OntClass restrictionClass = model.createClass(uri);
                restriction.addEquivalentClass(restrictionClass);
            }
            //createHasValueRestriction(java.lang.String uri, Property prop, RDFNode value)  
        }
    }
    
    /**
     * Generates the enumerated classes.
     */
    public void createEnumeratedClasses(){
        for(int i = 0; i < nrOfEnumClasses; i++){
            String uri = baseURI + "enumeratedClass" + i;
            ExtendedIterator eit = model.listIndividuals();
            OntoList instances = new OntoList();
            instances.addAll(eit.toList());
            int upper = 0;
            if(instances.size() < 100)
                upper = (int) Math.ceil(instances.size()/10);
            else 
                upper = (int) Math.ceil(instances.size()/100);
            int nrOfInstances = pseudoRand.nextInt(upper);
            RDFList members = model.createList();
            for(int j=0; j < nrOfInstances; j++){
                int member = pseudoRand.nextInt(instances.size());
                members = members.cons(instances.getIndividual(j));
            }
            OntClass enumClass = model.createEnumeratedClass(uri, members);
            connectingClasses.add(enumClass);
        }
    }
    
    /////-------------------------Getters/Setters-------------------------/////
    
    /**
     * Get some feedback about the generated class hierarchie of the ontology.
     * E.g. the number of classes, the number of subclasses and so on.
     *
     * @see ClassInformation
     * @return ClassInformation: an object containing all the information about the
     *                          generated classes.
     */
    public ClassInformation getClassInformation(){
        return classInformation;
    }
    
    /**
     * Gets the object properties used to create the restriction classes.
     *
     * @return ArrayList<OntProperty>: the object properties 
     *                                 used to create the 
     *                                 restriction classes.
     */
    public ArrayList<OntProperty> getRestrictionProps() {
        return restrictionProps;
    }
    
    /**
     * Gets the classes for which we should generate properties.
     *
     * @return ArrayList<OntClass>: the classes for which we generate properties.
     */
    public ArrayList<OntClass> getConnectingClasses() {
        return connectingClasses;
    }
}
