/*
 * Copyright 2008-2011 Femke Ongenae, Stijn Verstichel
 * distributed under the terms of the AGPL License
 * 
 * This file is part of OTAGen.
 * 
 * OTAGen is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Femke Ongenae & Stijn Verstichel
 * Ghent University - INTEC - IBCN
 * Zuiderpoort Block C0 bus 201
 * Gaston Crommenlaan 8
 * B-9050 Ghent
 * Belgium
 * E-mail: Femke.Ongenae@intec.ugent.be
 * E-mail: Stijn.Verstichel@intec.ugent.be
 * 
 */

package otagen;

import com.hp.hpl.jena.ontology.DatatypeProperty;
import com.hp.hpl.jena.ontology.OntClass;
import com.hp.hpl.jena.ontology.OntModel;
import com.hp.hpl.jena.ontology.OntResource;
import com.hp.hpl.jena.util.iterator.ExtendedIterator;
import com.hp.hpl.jena.vocabulary.XSD;
import java.util.ArrayList;
import java.util.Random;
import otagen.feedback.DatatypeInformation;

/**
 * Generates the datatype properties (+hierarchie) of the ontology.
 *
 * @author Femke Ongenae.
 * @author Stijn Verstichel.
 */
public class DataPropsGen {
    
    /////-------------------------FIELDS-------------------------/////
    
    //The base URI of the ontology
    private static String baseURI = null;
    
    //the generated ontology
    private OntModel model = null;
    
    //a pseudo random generator (with seed);
    private Random pseudoRand = null;
    
    //the number of classes
    private int nrOfClasses = 0;
    
    //the maximal depth of the classes and properties
    private int maxDepth = 0;
    
    //the maximal number of datatype properties that may be generated.
    private int nrOfDataProps = 0;
    
    //the maximal percentage of object properties (relations) or data properties with
    //complicated domain/range specifications.
    private int maxPercComplProperties = 0;
    
    //the maximal number of subproperties a property can have.
    private int maxNrSubProperties = 0;
    
    //holds the information about the generated datatype properties.
    private DatatypeInformation datatypeInformation = null;
    
    //holds the classes for which properties should be generated.
    private ArrayList<OntClass> connectingClasses = null;
    
    /////-------------------------Constructor-------------------------/////
    
    /** 
     * Creates the object properties (relations) (+hierarchie) of the ontology.
     *
     * @param OntModel model: the ontology we're creating.
     * @param String baseURI: the base URI of the ontology.
     * @param Random pseudoRand: a pseudo random generator (with seed).
     * @param int maxDepth: the maximal depth of the classes and properties.
     * @param int maxNrPercProperties: the maximal percentage of object properties (relations) or 
     *                                  data properties with complicated domain/range specifications.
     * @param int maxNrSubProperties: the maximal number of subproperties a property can have.
     * @param ArrayList<OntClass> connectingClasses: holds the classes for which properties should be generated.
     */
    public DataPropsGen(OntModel model, String baseURI, Random pseudoRand, int maxDepth, 
                        int maxPercComplProperties, int maxNrSubProperties, ArrayList<OntClass> connectingClasses) {
        //initialize fields
        this.model = model;
        this.baseURI = baseURI;
        this.pseudoRand = pseudoRand;
        this.maxDepth = maxDepth;
        this.maxPercComplProperties = maxPercComplProperties;
        this.maxNrSubProperties = maxNrSubProperties;
        this.connectingClasses = connectingClasses;
        datatypeInformation = new DatatypeInformation();
        
        this.nrOfClasses = connectingClasses.size();
        
        //The number of data properties is a pseudo random number.
        //The number is dependent on the nrOfClasses (more classes usually have
        //more data type properties).
        //It's unlikely to have more than 100 data properties.
        int limit = (int) Math.ceil(Math.min(nrOfClasses,100));
        nrOfDataProps = pseudoRand.nextInt(limit);
        if(nrOfDataProps == 0)
            nrOfDataProps = 1;
        datatypeInformation.setNrOfProps(nrOfDataProps);
        //Counter that counts the number of datatype properties already generated.
        int datCounter = 0;
        
        while(datCounter < nrOfDataProps)
        {
            int depth = 0;
            datCounter = createDataProperty(datCounter, null, depth);
        }
    }
    
    /////-------------------------Business Methods-------------------------/////
    
    /**
     * Method that is recursively used to generate datatype properties and their subproperties.
     *
     * @param int datCounter: counter that counts the number of datatype properties already generated.
     * @param int depth: counter that counts at which depth (in the subproperty tree) 
     *                   this property is generated.
     * @param DatatypeProperty superProp: -> null : this property doesn't have a superproperty 
     *                                              (is at the highest level).
     *                                    -> != null : the superproperty of the datatype property we're 
     *                                                 currently creating.
     * @return int: the number of datatype properties we have already created.
     */ 
    private int createDataProperty(int datCounter, DatatypeProperty superProp, int depth){
        if(datCounter < nrOfDataProps){
            if(datatypeInformation.getMaxDepthProps() < depth)
                datatypeInformation.setMaxDepthProps(depth);
            DatatypeProperty dat = model.createDatatypeProperty(baseURI + "DataProperty" + datCounter);
            datCounter++;
                
            ExtendedIterator superDom = null;
            ArrayList<OntClass> domainOptions = null;
            int domainOptionsSize = 0;
            if(superProp != null)
            {
                superProp.addSubProperty(dat);
                
                //For the domain
                //Only the classes that are part of the domain of the superproperty
                //and their subclasses can be part of the domain.
                superDom = superProp.listDomain();
                domainOptions = new ArrayList<OntClass>();
                while(superDom.hasNext())
                {
                    OntClass domain = (OntClass) superDom.next();
                    domainOptions.add(domain);
                    ExtendedIterator subDomains = domain.listSubClasses();
                    while(subDomains.hasNext())
                    {
                        OntClass subDomain = (OntClass) subDomains.next();
                        domainOptions.add(subDomain);
                    }
                }
                domainOptionsSize = domainOptions.size();
                
                //For the range
                //The subProperty has the same range as the superproperty
                OntResource range = superProp.getRange();
                dat.addRange(range);
                if(range.equals(XSD.xstring))
                    datatypeInformation.addRangeStringDataProp();
                else
                    datatypeInformation.addRangeIntegerDataProp();
            }
            else{
                //they all get range String or int as the specific range is not important.
                int range = pseudoRand.nextInt(2);
                //0: range is string, 1: range is integer
                if(range == 0)
                {
                    dat.addRange(XSD.xstring);
                    datatypeInformation.addRangeStringDataProp();
                }
                else
                {
                    dat.addRange(XSD.xint);
                    datatypeInformation.addRangeIntegerDataProp();
                }
            }
            
            //Does this datatype property have a complicated domain?
            //chance 1 in 4
            int hasComplDomain = pseudoRand.nextInt(4);
            //0: has complicated domain, 1,2,3: simple domain (one class)
            int nrOfDomainClasses = 1;
            //Number of datatype properties with complicated domains we may currently create
            int maxNrComplProp = (maxPercComplProperties*datCounter)/100;
            int nrOfComplProps = datatypeInformation.getNrOfComplProps();
            //Are we allowed to create more datatype properties with complicated domains
            if(hasComplDomain == 0 && nrOfComplProps < maxNrComplProp){
                boolean complicated = true; 
                if(superProp != null){
                    if(domainOptionsSize > 1){
                        int random = pseudoRand.nextInt(domainOptionsSize);
                        if(random > 1)
                            nrOfDomainClasses = random;
                        else
                            nrOfDomainClasses = 2;
                    }
                    else{
                        complicated = false;
                    }
                }
                else{
                    //does it have 2,3 or 4 classes as the domain?
                    nrOfDomainClasses = pseudoRand.nextInt(3) + 2;
                }
                if(complicated)
                    datatypeInformation.setNrOfComplProps(nrOfComplProps++);  
            }
            
            //Holds the classes we've allready added as a domain.
            //Used to check if we don't add the same class as domain twice.
            ArrayList<OntClass> domains = new ArrayList<OntClass>();
            //If it's not a subproperty we can choose random
            //classes to be part of the domain
            if(superProp == null)
            {
                for(int i = 0; i < nrOfDomainClasses; i++)
                {
                    //Check if this class isn't allready in the domain.
                    OntClass domainClass = null;
                    while(domains.contains(domainClass) || domainClass == null)
                    {
                        //Generate a random class as the domain class
                        int domain = pseudoRand.nextInt(nrOfClasses);
                        domainClass = connectingClasses.get(domain);
                    }
                    //add the class to the domain of the datatype property
                    dat.addDomain(domainClass);
                    domains.add(domainClass);
                }
            }
            //If it is a subproperty, then its domain is restricted
            //by the domain of his parent
            else
            {
                for(int i = 0; i < nrOfDomainClasses; i++)
                {
                    //Check if this class isn't allready in the domain.
                    OntClass domainClass = null;
                    while(domains.contains(domainClass) || domainClass == null)
                    {
                        //Generate a random class of the possible options as the domain class
                        int domain = pseudoRand.nextInt(domainOptions.size());
                        domainClass = domainOptions.get(domain);
                    }
                    //add the class to the domain of the datatype property
                    dat.addDomain(domainClass);
                    domains.add(domainClass);
                }
            }
            
            //is this a functional property?
            int isFunctional = pseudoRand.nextInt(2);
            //0 = functional, 1 = not functional
            if(isFunctional == 0){
                dat.convertToFunctionalProperty();
                datatypeInformation.addFunctionalDataProp();
            }
            
            //Does this class have subprops?
            //having subprops is less likely than having subclasses.
            //chance of having a subproperty is 1 in 4
            int hasSubprops = pseudoRand.nextInt(4);
            //0 = has subprops, 1,2,3 = no subprops
            //we can't exceed the maximal allowed depth of the ontology
            if(hasSubprops == 0 && depth != maxDepth)
            {
                //how many subprops does this class try to have?
                int nrOfSubProps = pseudoRand.nextInt(maxNrSubProperties);
                for(int i = 0; i < nrOfSubProps; i++)
                {
                    datCounter = createDataProperty(datCounter, dat, depth++);
                }
            }
        }
        return datCounter;
    }
    
    /////-------------------------Getters/Setters-------------------------/////
    
    /**
     * Get some feedback about the generated datatype property hierarchie of the ontology.
     * E.g. the number of datatype properties, the number of subproperties and so on.
     *
     * @see DatatypeInformation
     * @return DatatypeInformation: an object containing all the information about the
     *                              generated datatype properties.
     */
    public DatatypeInformation getDatatypeInformation(){
        return datatypeInformation;
    }
}
