/*
 * Copyright 2008-2011 Femke Ongenae, Stijn Verstichel
 * distributed under the terms of the AGPL License
 * 
 * This file is part of OTAGen.
 * 
 * OTAGen is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Femke Ongenae & Stijn Verstichel
 * Ghent University - INTEC - IBCN
 * Zuiderpoort Block C0 bus 201
 * Gaston Crommenlaan 8
 * B-9050 Ghent
 * Belgium
 * E-mail: Femke.Ongenae@intec.ugent.be
 * E-mail: Stijn.Verstichel@intec.ugent.be
 * 
 */
package otagen;

import com.hp.hpl.jena.ontology.OntClass;
import com.hp.hpl.jena.ontology.OntModel;
import com.hp.hpl.jena.ontology.OntModelSpec;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map.Entry;
import java.util.Properties;
import java.util.Random;
import java.util.Set;
import otagen.datatypes.Cluster;
import otagen.datatypes.RelationCount;
import otagen.feedback.AboxInformation;
import otagen.feedback.ClassInformation;
import otagen.feedback.DatatypeInformation;
import otagen.feedback.ObjectInformation;
import otagen.feedback.QueryInformation;
import otagen.feedback.TboxInformation;

/**
 * Generates an ontology (the conceptual level and the instance level).
 * 
 * @author Femke Ongenae.
 * @author Stijn Verstichel.
 */
public class Main {

    /////-------------------------FIELDS-------------------------/////
    
    //the generated ontology
    private OntModel model = null;
    //the pseudo random generator (with SEED)
    private Random pseudoRand = null;
    //the feedback about the generated T-Box
    private TboxInformation tBoxFeedBack = null;
    //the feedback about the generated A-Box
    private AboxInformation aBoxFeedBack = null;
    //the feedback about the generated query-sets
    private QueryInformation queryFeedBack = null;

    /////-------------------------Constructor-------------------------/////
    
    /** 
     * Creates a new instance of Main.
     * Initializes the pseudo random generator that will be used.
     * @param int seed: The SEED that will be used to control the behaviour of the peudo random generator.
     *                  Each time you use the same seed, the same nunbers will be generated.
     */
    public Main(int seed) {
        //Initialize the pseudo random generator with the seed.
        //If the same seed is used, the same ontology will be generated.
        pseudoRand = new Random(seed);
    }

    /////-------------------------Busines methods-------------------------/////
    
    /**
     * Generates an ontology and it's instance data.
     *
     * Uses a properties file to determine the following properties:
     * 1. the number of classes 
     * 2. the maximal depth of the classes and properties
     * 3. the maximal number of subclasses a class can have
     * 4. A flag: true : create logical defined classes 
     *            false : create some empty (non-defined) classes 
     * 5. the number of classes of the previous category that must be created 
     * 6. the minimal connectivity of class-relations graph
     * 7. the maximal percentage of object properties (relations) or datatype properties with 
     *    complicated domain/range specifications 
     * 8. the SEED
     */
    public static void main(String[] args) {
        
        System.out.println("    OTAGen Copyright (C) 2008-2011 Femke Ongenae, Stijn Verstichel\n"
                + "    This program comes with ABSOLUTELY NO WARRANTY; \n"
                + "    This is free software, and you are welcome to redistribute it\n"
                + "    under conditions of the AGPL License.");
        
        try {
            // Read properties file.
            Properties properties = new Properties();
            File dir = new File(".");
            FileInputStream in = new FileInputStream(dir.getCanonicalPath() + "/conf/ontgen.properties");
            properties.load(in);
            in.close();

            // read the T-Box properties
            int nrOfClasses = Integer.parseInt(properties.getProperty("Number_of_classes"));
            int maxDepth = Integer.parseInt(properties.getProperty("Max_depth_tree"));
            int maxNrSubClasses = Integer.parseInt(properties.getProperty("Max_number_of_subclasses"));
            int maxNrSubProperties = Integer.parseInt(properties.getProperty("Max_number_of_subproperties"));
            boolean flag = Boolean.parseBoolean(properties.getProperty("Empty_class_flag"));
            int nrOfFlagClasses = Integer.parseInt(properties.getProperty("Number_of_logical_empty_classes"));
            int minConnectivity = Integer.parseInt(properties.getProperty("Min_connectivity"));
            int maxPercComplProperties = Integer.parseInt(
                    properties.getProperty("Max_percent_of_complicated_properties"));
            int seed = Integer.parseInt(properties.getProperty("Seed"));

            // read the A-Box properties
            int clusterConnectivity = Integer.parseInt(properties.getProperty("ClusterConnectivity"));
            int clusterCount = Integer.parseInt(properties.getProperty("ClusterCount"));
            int maxClusterSize = Integer.parseInt(properties.getProperty("MaxClusterSize"));
            int minClusterSize = Integer.parseInt(properties.getProperty("MinClusterSize"));
            int minIndividualCount = Integer.parseInt(properties.getProperty("MinIndividualCount"));
            int datatypePropertyCount = Integer.parseInt(properties.getProperty("DatatypePropertyCount"));
            int individualDatatypePropertyPercentage = Integer.parseInt(properties.getProperty("IndividualDatatypePropertyPercentage"));
            int minObjectRelationCount = Integer.parseInt(properties.getProperty("MinObjectPropertyCount"));
            int maxObjectRelationCount = Integer.parseInt(properties.getProperty("MaxObjectPropertyCount"));

            // read the Query properties
            int queryDepthCount = Integer.parseInt(properties.getProperty("queryDepthCount"));
            int[] queryDepths = new int[queryDepthCount];
            int[] queryCounts = new int[queryDepthCount];
            for (int i = 0; i < queryDepthCount; i++) {
                queryDepths[i] = Integer.parseInt(properties.getProperty("queryDepth" + (i + 1)));
                queryCounts[i] = Integer.parseInt(properties.getProperty("queryCountForDepth" + (i + 1)));
            }

            Main m = new Main(seed);

            // generate the ontology
            m.generateOntology(nrOfClasses, maxDepth, maxNrSubClasses, flag, nrOfFlagClasses,
                    minConnectivity, maxPercComplProperties, maxNrSubProperties,
                    clusterConnectivity, clusterCount, maxClusterSize,
                    minClusterSize, minIndividualCount, datatypePropertyCount,
                    individualDatatypePropertyPercentage, minObjectRelationCount, 
                    maxObjectRelationCount, queryDepths, queryCounts);

            System.out.println("[STATUS] Writing Knowledge Base to File");
            // Write to file
            m.writeOntology();

            System.out.println("[STATUS] =================================");
            // Provide Feedback
            m.provideFeedback(nrOfClasses, nrOfFlagClasses);

        } catch (Exception e) {
            System.err.println(e);
            e.printStackTrace();
        }
    }

    /**
     * Generates the ontology (the concept level and the data level).
     * 
     * @param int nrOfClasses: the number of classes.
     * @param int maxDepth: the maximal depth of the classes or properties.
     * @param maxNrSubClasses: the maximal number of subclasses a class can have.
     * @param boolean flag: A flag: true : create logical defined classes,
     *                              false : create some empty (non-defined) classes.
     * @param int nrOfFlagClasses: the number of classes of the previous category that must be created.
     * @param int minConnectivity: the minimal connectivity of class-relations graph.
     * @param int maxPercComplProperties: the maximal percentage of object properties (relations) 
     *                                  or datatype properties with complicated domain/range specifications.
     * @param int maxNrSubProperties: the maximal number of subproperties a property can have.
     * @param int clusterConnectivity: Specifies the percentage of fully connected instances amongst the 
     *                                 instances of the concepts in the cluster
     * @param int clusterCount: Specifies the amount of clusters to be generated in the Ontology Model
     * @param int minClusterSize: Specifies the minimum amount of concepts to be included in the cluster
     * @param int maxClusterSize: Specifies the maximum amount of concept to be included in a cluster
     * @param int minIndividualCount: Specifies the minimum amount of individuals to be generated
     * @param OntModel ontModel: The Jena internal OntModel, as constructed with the generated T-Box
     * @param Random pseudoRand: The seed-based random generator, the same as being used by the T-Box generator.
     * @param int datatypePropertyCount: indicates how many instances of the datatypeproperty are to be generated in the case the property is non-functional
     * @param int individualDatatypePropertyPercentage: indicates how many of the individuals should have their datatype properties filled in
     * @param int minObjectPropertyCount: Indicates minimum instances of the objectproperty are to be generated in the case the property is non-functional (intra-cluster)
     * @param int maxObjectPropertyCount: Indicates maximum instances of the objectproperty are to be generated in the case the property is non-functional (intra-cluster)* 
     * @param int[] queryDepths: specifies for every query-set to be generated the depths of the queries
     * @param int[] queryCounts: specifies for every query-set the be generated the amount of the queries
     */
    private void generateOntology(int nrOfClasses, int maxDepth, int maxNrSubClasses, boolean flag,
            int nrOfFlagClasses, int minConnectivity, int maxPercComplProperties,
            int maxNrSubProperties, int clusterConnectivity,
            int clusterCount, int maxClusterSize, int minClusterSize,
            int minIndividualCount, int datatypePropertyCount,
            int individualDatatypePropertyPercentage, int minObjectPropertyCount, 
            int maxObjectPropertyCount, int[] queryDepths, int[] queryCounts) {

        //create an empty model
        //model = ModelFactory.createOntologyModel(PelletReasonerFactory.THE_SPEC);
        model = ModelFactory.createOntologyModel(OntModelSpec.OWL_MEM);

        System.out.println("[STATUS] Generating T-Box");
        // generate T-Box
        TboxGen tBox = new TboxGen(nrOfClasses, maxDepth, maxNrSubClasses, flag, nrOfFlagClasses,
                minConnectivity, maxPercComplProperties, maxNrSubProperties,
                model, pseudoRand);
        // get information about the generated T-Box
        tBoxFeedBack = tBox.getFeedBack();

        System.out.println("[STATUS] Generating A-Box");
        // Generate A-box

        AboxGen aBox = new AboxGen(clusterConnectivity, clusterCount, maxClusterSize,
                minClusterSize, minIndividualCount, model, pseudoRand, datatypePropertyCount,
                individualDatatypePropertyPercentage, minObjectPropertyCount,
                maxObjectPropertyCount);
        // get information about the generated A-Box
        aBoxFeedBack = aBox.getFeedBack();

        // Generate Enumerated classes
        System.out.println("[STATUS] Generating Enumerated Classes");
        tBox.generateEnumClasses();
        
        // Generate object properties that fulfill the minimum connectivity of the
        // enumerated and restriction classes.
        System.out.println("[STATUS] Generating Object Properties for Enumerated Classes");
        tBox.connectEnumClasses();

        //After the domain and ranges have been set, the individuals of these
        //Enum classes have to be linked using these properties.
        System.out.println("[STATUS] Instantiating Object Properties for Enumerated Classes");
        aBox.generateEnumRelations();

        System.out.println("[STATUS] Generating Queries");
        // Generate Queries
        QueryGen queryGen = new QueryGen(model, pseudoRand, queryDepths,
                queryCounts);
        // get information about the generated Query-Sets
        queryFeedBack = queryGen.getFeedBack();
    }

    /**
     * Writes the created ontology in a file;
     *
     * @throws FileNotFoundException if the file to which is written cannot be found.
     */
    private void writeOntology() throws FileNotFoundException {

        // Create FileOutputStream
        FileOutputStream out = new FileOutputStream(new File("generated/generatedOnt.owl"));

        // write the ontology to the outputstream
        model.write(out, "RDF/XML-ABBREV");
    }

    /**
     * Provides feedback about the created ontology.
     * E.g. the number of object properties, data properties, ...
     * E.g. the number of instances of each class, ...
     */
    private void provideFeedback(int nrOfClasses, int nrOfFlagClasses) {
        System.out.println("\nFEEDBACK: \n");

        //ABOUT THE T-BOX
        System.out.println("FEEDBACK ABOUT THE T-BOX: \n");
        //ABOUT CLASSES
        System.out.println("FEEDBACK ABOUT THE CLASSES:");
        System.out.println("--------------------------- \n");
        ClassInformation classInformation = tBoxFeedBack.getClassInformation();
        System.out.println("Nr of Classes Created: " + nrOfClasses + " " + model.listNamedClasses().toList().size());
        System.out.println("Nr of Simple Classes Created: " + (nrOfClasses - nrOfFlagClasses));
        System.out.println("Nr of Logical Classes Created: " + nrOfFlagClasses);
        System.out.println("Nr of Empty Classes Created: " + classInformation.getNrOfEmptyClasses());
        System.out.println("Nr of Union Classes Created: " + classInformation.getNrOfUnionClasses() + " " + model.listUnionClasses().toList().size());
        System.out.println("Nr of Intersection Classes Created: " + classInformation.getNrOfIntersectionClasses() + " " + model.listIntersectionClasses().toList().size());
        System.out.println("Nr of Complement Classes Created: " + classInformation.getNrOfComplementClasses() + " " + model.listComplementClasses().toList().size());
        System.out.println("Nr of Enumerated Classes Created: " + classInformation.getNrOfEnumClasses() + " " + model.listEnumeratedClasses().toList().size());
        System.out.println("Nr of Restriction Classes Created: " + classInformation.getNrOfRestrictionClasses());
        System.out.println("Maximal Depth of the Simple (Not Logical) Classes: " + classInformation.getSimpleDepth());
        System.out.println("\n");

        //ABOUT OBJECT PROPERTIES (RELATIONS)
        System.out.println("FEEDBACK ABOUT THE OBJECT PROPERTIES: ");
        System.out.println("------------------------------------- \n");
        ObjectInformation objectInformation = tBoxFeedBack.getObjectInformation();
        System.out.println("Nr of Relations (Object Properties) Created: " + objectInformation.getNrOfProps() + " " + model.listObjectProperties().toList().size());
        System.out.println("Maximal Depth of the Relations: " + objectInformation.getMaxDepthProps());
        System.out.println("Nr of Relations with Complicated Domains/Ranges: " + objectInformation.getNrOfComplProps());
        System.out.println("Nr of Relations with Complicated Domains: " + objectInformation.getNrOfComplDomProps());
        System.out.println("Nr of Relations with Complicated Ranges: " + objectInformation.getNrOfComplRangeProps());
        System.out.println("Nr of Functional Relations Created: " + objectInformation.getNrOfFunctionalProps() + " " + model.listFunctionalProperties().toList().size());
        System.out.println("Nr of Transitive Relations Created: " + objectInformation.getNrOfTransitiveProps() + " " + model.listTransitiveProperties().toList().size());
        System.out.println("Nr of Symmetric Relations Created: " + objectInformation.getNrOfSymmetricProps() + " " + model.listSymmetricProperties().toList().size());
        System.out.println("Nr of Inverse Relations Created: " + objectInformation.getNrOfInverseProps() + " " + model.listInverseFunctionalProperties().toList().size());
        System.out.println("\n");
        System.out.println("The Connectivity of Each Class: ");
        System.out.println("------------------------------- \n");
        HashMap<OntClass, Integer> connect = objectInformation.getConnect();
        Set<OntClass> keys = connect.keySet();
        Iterator<OntClass> it = keys.iterator();
        while (it.hasNext()) {
            OntClass key = it.next();
            System.out.println("Class: " + key);
            System.out.println("Connectivity: " + connect.get(key));
        }
        System.out.println("\n");

        //ABOUT DATATYPE PROPERTIES
        System.out.println("FEEDBACK ABOUT THE DATATYPE PROPERTIES: ");
        System.out.println("--------------------------------------- \n");
        DatatypeInformation datatypeInformation = tBoxFeedBack.getDatatypeInformation();
        System.out.println("Nr of Datatype Properties Created: " + datatypeInformation.getNrOfProps() + " " + model.listDatatypeProperties().toList().size());
        System.out.println("Maximal Depth of the Datatype Properties: " + datatypeInformation.getMaxDepthProps());
        System.out.println("Nr of Datatype Properties with Complicated Domains: " + datatypeInformation.getNrOfComplProps());
        System.out.println("Nr of Functional Datatype Properties: " + datatypeInformation.getNrOfFunctionalProps() + " " + model.listFunctionalProperties().toList().size());
        System.out.println("Nr of Datatype Properties with Range String: " + datatypeInformation.getNrOfRangeStringProps());
        System.out.println("Nr of Datatype Properties with Range int: " + datatypeInformation.getNrOfRangeIntProps());
        System.out.println("\n");

        //ABOUT THE A-BOX
        System.out.println("FEEDBACK ABOUT THE A-BOX: \n ");
        System.out.println("FEEDBACK ABOUT THE CLUSTERS: ");
        System.out.println("---------------------------- \n");

        System.out.println("Nr of Clusters Created: " + aBoxFeedBack.getGeneratedClusters().size());
        Iterator<Cluster> cit = aBoxFeedBack.getGeneratedClusters().iterator();
        int count = 0;
        while (cit.hasNext()) {
            System.out.println("Cluster nr " + (count + 1) + ": " + cit.next());
            count++;
        }

        System.out.println("\n");
        System.out.println("FEEDBACK ABOUT THE INSTANCES OF CLASSES: ");
        System.out.println("---------------------------------------- \n");

        Set<Entry<OntClass, Integer>> values = aBoxFeedBack.getIndividualCount().
                getGeneratedIndividuals().entrySet();
        Iterator<Entry<OntClass, Integer>> iit = values.iterator();
        int sum = 0;
        Entry<OntClass, Integer> entry = null;
        while (iit.hasNext()) {
            entry = iit.next();
            System.out.println("Instances for: " + entry.getKey() + " : " +
                    entry.getValue());
            sum += entry.getValue().intValue();
        }
        System.out.println("Nr of Instances of Classes Created: " + sum);

        System.out.println("\n");
        System.out.println("FEEDBACK ABOUT THE INSTANCES OF OBJECT PROPERTIES: ");
        System.out.println("-------------------------------------------------- \n");
        Iterator<RelationCount> rcit = aBoxFeedBack.getRelationCount().iterator();
        RelationCount rc = null;
        count = 0;
        while (rcit.hasNext()) {
            rc = rcit.next();
            if (rc.getRange() != null) {
                System.out.println("Nr of Instances of Relation: " +
                        rc.getProperty() + " between Domain: " +
                        rc.getDomain() + " and Range: " + rc.getRange() +
                        ": " + rc.getCount());
                count += rc.getCount();
            }
        }
        System.out.println("Nr (Total) of Instances of Relations Created: " +
                count);

        System.out.println("\n");
        System.out.println("FEEDBACK ABOUT THE INSTANCES OF DATATYPE PROPERTIES: ");
        System.out.println("---------------------------------------------------- \n");
        count = 0;
        Iterator<RelationCount> rcitr = aBoxFeedBack.getRelationCount().iterator();
        while (rcitr.hasNext()) {
            rc = rcitr.next();
            if (rc.getRangeR() != null) {
                System.out.println("Nr of Instances of Relation: " +
                        rc.getProperty() + " between Domain: " +
                        rc.getDomain() + " and Range: " + rc.getRangeR() +
                        ": " + rc.getCount());
                count += rc.getCount();
            }
        }
        System.out.println("Nr of Instances of Datatype Properties Created: " + count);

        System.out.println("\n");
        System.out.println("FEEDBACK ABOUT THE QUERIES: ");
        System.out.println("------------------------- \n");

        Iterator<String> qit = this.queryFeedBack.getQueries().iterator();
        count = 0;
        while (qit.hasNext()) {
            System.out.println("Query nr: " + count + " \n " + qit.next());
            count++;
            System.out.println("++++++++++++++++++++++++++++++++++++++++++++++");
        }
    }
}
