/*
 * Copyright 2008-2011 Femke Ongenae, Stijn Verstichel
 * distributed under the terms of the AGPL License
 * 
 * This file is part of OTAGen.
 * 
 * OTAGen is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Femke Ongenae & Stijn Verstichel
 * Ghent University - INTEC - IBCN
 * Zuiderpoort Block C0 bus 201
 * Gaston Crommenlaan 8
 * B-9050 Ghent
 * Belgium
 * E-mail: Femke.Ongenae@intec.ugent.be
 * E-mail: Stijn.Verstichel@intec.ugent.be
 * 
 */

package otagen;

import com.hp.hpl.jena.ontology.ObjectProperty;
import com.hp.hpl.jena.ontology.OntClass;
import com.hp.hpl.jena.ontology.OntModel;
import com.hp.hpl.jena.ontology.OntProperty;
import com.hp.hpl.jena.util.iterator.ExtendedIterator;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Random;
import java.util.Set;
import otagen.datatypes.OntoList;
import otagen.feedback.ObjectInformation;

/**
 * Generates the object properties (relations) (+hierarchie) of the ontology.
 *
 * @author Femke Ongenae.
 * @author Stijn Verstichel.
 */
public class ObjPropsGen {
    
    /////-------------------------FIELDS-------------------------/////
    
    //The base URI of the ontology
    private static String baseURI = null;
    
    //the generated ontology
    private OntModel model = null;
    
    //a pseudo random generator (with seed);
    private Random pseudoRand = null;

    //the minimal connectivity of class-relations graph
    private int minConnectivity = 0;
    
    //the maximal depth of the classes and properties
    private int maxDepth = 0;
    
    //the maximal percentage of object properties (relations) or data properties with
    //complicated domain/range specifications.
    private int maxPercComplProperties = 0;
    
    //the maximal number of subproperties a property can have.
    private int maxNrSubProperties = 0;
    
    //holds the information about the generated object properties.
    private ObjectInformation objectInformation = null;

    //holds the number of enumerated classes that will be created.
    private int nrOfEnumClasses;
    
    //holds the number of restriction classes that will be created.
    private int nrOfRestrictionClasses;
    
    //Maps each class on its connectivity.
    private HashMap<OntClass,Integer> connect;
    
    //holds the object properties used to create the restriction classes.
    private ArrayList<OntProperty> restrictionProps = null;
    
    //holds the classes for which properties should be generated.
    private ArrayList<OntClass> connectingClasses = null;
    
    /////-------------------------Constructor-------------------------/////
    
    /** 
     * Creates the object properties (relations) (+hierarchie) of the ontology.
     *
     * @param OntModel model: the ontology we're creating.
     * @param String baseURI: the base URI of the ontology.
     * @param Random pseudoRand: a pseudo random generator (with seed).
     * @param int maxDepth: the maximal depth of the classes and properties.
     * @param int minConnectivity: the minimal connectivity of class-relations graph.
     * @param int maxNrPercProperties: the maximal percentage of object properties (relations) or 
     *                                  data properties with complicated domain/range specifications.
     * @param int maxNrSubProperties: the maximal number of subproperties a property can have.
     * @param ArrayList<OntClass> connectingClasses: holds the classes for which properties should be generated.
     */
    public ObjPropsGen(OntModel model, String baseURI, Random pseudoRand, int maxDepth, 
                       int minConnectivity, int maxPercComplProperties, int maxNrSubProperties,
                       int nrOfEnumClasses, int nrOfRestrictionClasses,
                       ArrayList<OntClass> connectingClasses) {
        //initialize fields
        this.model = model;
        this.baseURI = baseURI;
        this.pseudoRand = pseudoRand;
        this.maxDepth = maxDepth;
        this.minConnectivity = minConnectivity;
        this.maxPercComplProperties = maxPercComplProperties;
        this.maxNrSubProperties = maxNrSubProperties;
        this.nrOfEnumClasses = nrOfEnumClasses;
        this.nrOfRestrictionClasses = nrOfRestrictionClasses;
        this.connectingClasses = connectingClasses;
        objectInformation = new ObjectInformation();
        
        int nrOfClasses = connectingClasses.size();
        connect = new HashMap<OntClass,Integer>();
        for(int i = 0; i < nrOfClasses; i++)
            connect.put(connectingClasses.get(i),new Integer(0));
        
        //Compute how many object properties we must create at least to meet
        //the connectivity demands
        double minNrObjProps = Math.ceil((nrOfClasses * minConnectivity) / 2.0);
        //Counts how many object properties have allready been created.
        int objCounter = 0;
        //now randomly create these object properties
        while(objCounter < minNrObjProps)
        {
            int depth = 0;
            objCounter = createObjProperty(objCounter,null,depth,minNrObjProps,connect);
        }
        
        //Some classes don't have the minimum connectivity yet because
        //the relations were randomly generated.  Create extra object
        //properties so that these classes also fulfill the minimum connectivity.
        ArrayList<OntClass> notConnected = searchNotConnected(connect);
        //If all classes have the minimum connectivity we're done generating relations.
        while(notConnected.size() != 0){
            //Create an object property that takes the two classes, that don't have
            //minimum connectivity yet, as domain and range.
            ObjectProperty obj = model.createObjectProperty(baseURI + "ObjectProperty" + objCounter);
            
            //Randomly choose a class that doesn't have minimum connectivity yet
            //as the domain.
            int chooseDomain = pseudoRand.nextInt(notConnected.size());
            OntClass domain = notConnected.get(chooseDomain);
            obj.setDomain(domain);  
            
            //Check if there's still a second class left that doesn't have
            //the minimum connectivity yet and hasn't been chosen as the domain yet.
            OntClass range = null;
            if(notConnected.size() != 1){
                //Randomly choose a class that doesn't have minimum connectivity yet
                //as the range (has to be a different class than the domain class, 
                //because loops do not count towards the connectivity).
                while(range == domain || range == null){
                    int chooseRange = pseudoRand.nextInt(notConnected.size());
                    range = notConnected.get(chooseRange);
                }    
            }
            else{
                //Randomly choose a class as the range(has to be a different class 
                //than the domain class, because loops do not count towards the connectivity).
                while(range == domain || range == null){
                    int chooseRange = pseudoRand.nextInt(nrOfClasses);
                    range = connectingClasses.get(chooseRange);
                }      
            }
            obj.setRange(range);
            
            //adjust the connected table for the range
            Integer oldConnect = connect.get(range);
            Integer newConnect = new Integer(oldConnect.intValue() + 1);
            connect.put(range,newConnect); 
            if(notConnected.size() != 1){
                //Check if the connectivity demands for this class are reached.
                if(newConnect >= minConnectivity)
                    notConnected.remove(range);
            }
            
            //adjust the connected table for the domain
            oldConnect = connect.get(domain);
            newConnect = new Integer(oldConnect.intValue() + 1);
            connect.put(domain,newConnect); 
            //Check if the connectivity demands for this class are reached.
            if(newConnect >= minConnectivity)
                notConnected.remove(domain);
            
            //Configure the characteristics of the relation e.g. transitivity
            addFuncSymm(obj, null);
            addTransitive(obj, null);
        }
        
        //create object properties for the enumerated 
        //classes that will be generated later
        objCounter = createEnumObjProp(objCounter);
        
        //configure feedback about the generation of the relations.
        objectInformation.setNrOfProps(objCounter);
    }

    /////-------------------------Business Methods-------------------------/////
    
    /**
     * Method that is recursively used to generate object properties and their subproperties.
     *
     * @param int objCounter: counter that counts the number of object properties already generated.
     * @param int depth: counter that counts at which depth (in the subproperty tree) 
     *                   this property is generated.
     * @param ObjectProperty superProp: -> null : this property doesn't have a superproperty 
     *                                            (is at the highest level).
     *                                  -> != null : the superproperty of the object property we're 
     *                                               currently creating.
     * @param OntoList classes: All the classes that are contained in the ontology.
     * @param double minNrObjProps: the minimal number of object properties that have to be created
     *                              to meet the connectivity demands.
     * @return int: the number of object properties we have already created.
     */ 
    private int createObjProperty(int objCounter, ObjectProperty superProp, int depth, 
                                  double minNrObjProps, HashMap<OntClass,Integer> connect){
        if(objCounter < minNrObjProps){
            if(objectInformation.getMaxDepthProps() < depth)
                objectInformation.setMaxDepthProps(depth);
            ObjectProperty obj = model.createObjectProperty(baseURI + "ObjectProperty" + objCounter);
            objCounter++;
              
            //Configures if this property has an inverse property
            int hasInverse = pseudoRand.nextInt(5);
            //0 = has inverse, 1,2,3,4 = not has inverse
            ObjectProperty inverse = null;
            if(hasInverse == 0){
                inverse = model.createObjectProperty(baseURI + "ObjectProperty" + objCounter);
                objCounter++;
                objectInformation.addInverseProp();
                inverse.setInverseOf(obj);
            }
            
            //Configure superProperty
            ExtendedIterator superDom = null;
            ArrayList<OntClass> domainOptions = null;
            ExtendedIterator superRange = null;
            ArrayList<OntClass> rangeOptions = null;
            int domainOptionsSize = 0;
            int rangeOptionsSize = 0;
            if(superProp != null)
            {
                superProp.addSubProperty(obj);
                
                //For the domain
                //Only the classes that are part of the domain of the superproperty
                //and their subclasses can be part of the domain.
                superDom = superProp.listDomain();
                domainOptions = new ArrayList<OntClass>();
                while(superDom.hasNext())
                {
                    OntClass domain = (OntClass) superDom.next();
                    domainOptions.add(domain);
                    ExtendedIterator subDomains = domain.listSubClasses();
                    while(subDomains.hasNext())
                    {
                        OntClass subDomain = (OntClass) subDomains.next();
                        domainOptions.add(subDomain);
                    }
                }
                
                //For the range
                //Only the classes that are part of the range of the superproperty
                //and their subclasses can be part of the domain.
                superRange = superProp.listRange();
                rangeOptions = new ArrayList<OntClass>();
                while(superRange.hasNext())
                {
                    OntClass range = (OntClass) superRange.next();
                    rangeOptions.add(range);
                    ExtendedIterator subRanges = range.listSubClasses();
                    while(subRanges.hasNext())
                    {
                        OntClass subRange = (OntClass) subRanges.next();
                        rangeOptions.add(subRange);
                    }
                }
                
                domainOptionsSize = domainOptions.size();
                rangeOptionsSize = rangeOptions.size();
            }
            
            //Does this datatype property have a complicated domain or range?
            //chance 1 in 4
            int hasComplDomain = pseudoRand.nextInt(4);
            int hasComplRange = pseudoRand.nextInt(4);
            //0: has complicated domain/range, 1,2,3: simple domain/range (one class)
            int nrOfDomainClasses = 1;
            int nrOfRangeClasses = 1;
            //Number of object properties with complicated domains/ranges we may currently create
            int maxNrComplProp = (maxPercComplProperties*objCounter)/100;
            int nrOfComplProps = objectInformation.getNrOfComplProps();
            //Are we allowed to create more datatype properties with complicated domains
            if((hasComplDomain == 0 || hasComplRange == 0) && nrOfComplProps < maxNrComplProp){
                boolean complicated = true; 
                if(hasComplDomain == 0){
                    if(superProp != null){
                        if(domainOptionsSize > 1){
                            int random = pseudoRand.nextInt(domainOptionsSize);
                            if(random > 1)
                                nrOfDomainClasses = random;
                            else
                                nrOfDomainClasses = 2;
                            objectInformation.addComplDomProp();
                        }
                        else
                            complicated = false;
                    }
                    else{
                        nrOfDomainClasses = pseudoRand.nextInt(3) + 2;
                        objectInformation.addComplDomProp();
                    }
                 }
                else{
                    if(superProp != null){
                        if(rangeOptionsSize > 1){
                            int random = pseudoRand.nextInt(rangeOptionsSize);
                            if(random > 1)
                                nrOfRangeClasses = random;
                            else
                                nrOfRangeClasses = 2;
                            objectInformation.addComplRangeProp();
                        }
                        else{
                            complicated = false;
                        }
                    }
                    else{
                        nrOfRangeClasses = pseudoRand.nextInt(3) + 2;
                        objectInformation.addComplRangeProp();
                    }
                }
                if(complicated)
                    objectInformation.setNrOfComplProps(nrOfComplProps++);
            }
            
            //List that keep track of the domains and ranges 
            //that are chosen. Used to update the connected 
            //map (see further) and to check if the same classes
            //isn't added to the domain/range twice.
            ArrayList<OntClass> domains = new ArrayList<OntClass>();
            ArrayList<OntClass> ranges = new ArrayList<OntClass>();
            
            //If it's not a subproperty we can choose random
            //classes to be part of the domain/range
            if(superProp == null)
            {
                for(int i = 0; i < nrOfDomainClasses; i++)
                {
                    //Check if this class isn't allready in the domain.
                    OntClass domainClass = null;
                    while(domains.contains(domainClass) || domainClass == null)
                    {
                        //Generate a random class as the domain class
                        int domain = pseudoRand.nextInt(connectingClasses.size());
                        domainClass = connectingClasses.get(domain);
                    }
                    //add the class to the domain of the object property
                    obj.addDomain(domainClass);
                    if(hasInverse == 0)
                        inverse.addRange(domainClass);
                    domains.add(domainClass);
                }
                for(int i = 0; i < nrOfRangeClasses; i++)
                {
                    //Check if this class isn't allready in the range.
                    OntClass rangeClass = null;
                    while(ranges.contains(rangeClass) || rangeClass == null)
                    {
                        //Generate a random class as the range class
                        int range = pseudoRand.nextInt(connectingClasses.size());
                        rangeClass = connectingClasses.get(range);
                    }
                    //add the class to the range of the object property
                    obj.addRange(rangeClass);
                    if(hasInverse == 0)
                        inverse.addDomain(rangeClass);
                    ranges.add(rangeClass);
                }
            }
            //If it is a subproperty, then its domain/range is restricted
            //by the domain/range of his parent
            else
            {
                //For the domain
                for(int i = 0; i < nrOfDomainClasses; i++)
                {
                    //Check if this class isn't allready in the domain.
                    OntClass domainClass = null;
                    while(domains.contains(domainClass) || domainClass == null)
                    {
                        //Generate a random class of the possible options as the domain class
                        int domain = pseudoRand.nextInt(domainOptions.size());
                        domainClass = domainOptions.get(domain);
                    }
                    //add the class to the domain of the object property
                    obj.addDomain(domainClass);
                    if(hasInverse == 0)
                        inverse.addRange(domainClass);
                    domains.add(domainClass);
                }
                 
                //For the range
                for(int i = 0; i < nrOfRangeClasses; i++)
                {
                    //Check if this class isn't allready in the range.
                    OntClass rangeClass = null;
                    while(ranges.contains(rangeClass) || rangeClass == null)
                    {
                        //Generate a random class of the possible options as the range class
                        int range = pseudoRand.nextInt(rangeOptions.size());
                        rangeClass = rangeOptions.get(range);
                    }
                    //add the class to the domain of the object property
                    obj.addRange(rangeClass);
                    if(hasInverse == 0)
                        inverse.addDomain(rangeClass);
                    ranges.add(rangeClass);
                }
            }
            
            //adjust the connected table
            //Each domain class can be connected with each range class,
            //so it's connectivity raises with the number of range classes.
            //EXCEPTION: Loops (a connections between a domain and a range that is the 
            //same class) do not count towards the connectivity demands.
            for(int i = 0; i < nrOfDomainClasses; i++){
                OntClass domain = domains.get(i);
                Integer oldConnect = connect.get(domain);
                Integer newConnect = new Integer(oldConnect.intValue() + nrOfRangeClasses);
                if(hasInverse == 0)
                    newConnect += nrOfRangeClasses;
                for(int j = 0; j < nrOfRangeClasses; j++){
                    OntClass range = ranges.get(j);
                    if(domain == range)
                        newConnect--;
                    if(hasInverse == 0)
                        newConnect--;
                }
                connect.put(domain,newConnect);
            } 
            //Each range class can be connected with each domain class,
            //so it's connectivity raises with the number of range classes.
            //EXCEPTION: Loops (a connections between a domain and a range that is the 
            //same class) do not count towards the connectivity demands.
            for(int i = 0; i < nrOfRangeClasses; i++){
                OntClass range = ranges.get(i);
                Integer oldConnect = connect.get(range);
                Integer newConnect = new Integer(oldConnect.intValue() + nrOfDomainClasses);
                if(hasInverse == 0)
                    newConnect += nrOfDomainClasses;
                for(int j = 0; j < nrOfDomainClasses; j++){
                    OntClass domain = domains.get(j);
                    if(domain == range)
                        newConnect--;
                    if(hasInverse == 0)
                        newConnect--;
                }
                connect.put(range,newConnect);
            }
            
            //Configure the properties of this object property.
            //E.g. Transitivity, Symmetric, Functionality,...
            addFuncSymm(obj, inverse);
            addTransitive(obj, inverse);
            
            //Does this class have subprops?
            //having subprops is less likely than having subclasses.
            //chance of having a subproperty is 1 in 4
            int hasSubprops = pseudoRand.nextInt(4);
            //0 = has subprops, 1,2,3 = no subprops
            //we can't exceed the maximal allowed depth of the ontology
            if(hasSubprops == 0 && depth != maxDepth)
            {
                //how many subprops does this class try to have?
                int nrOfSubProps = pseudoRand.nextInt(maxNrSubProperties);
                for(int i = 0; i < nrOfSubProps; i++)
                {
                    objCounter = createObjProperty(objCounter, obj, depth++, minNrObjProps, connect);
                }
            }
        }
        return objCounter;
    }
    
    /**
     * whether an object property is functional/symmetric.
     *
     * @param ObjectProperty obj: the object property of which we 
     *                            want to the characteristics.
     * @param ObjectProperty inverse: the inverse object property that 
     *                                needs characteristics that correspond
     *                                with the property of which it's the inverse.
     */
    private void addFuncSymm(ObjectProperty obj, ObjectProperty inverse){
        
        //is this a functional property?
        int isFunctional = pseudoRand.nextInt(3);
        //0 = functional, 1,2 = not functional
        if(isFunctional == 0){
            obj.convertToFunctionalProperty();
            objectInformation.addFunctionalProp();
        }
            
        //is this a symmetric property?
        int isSymmetric = pseudoRand.nextInt(3);
        //0 = symmetric, ,2 = not symmetric
        if(isSymmetric == 0){
            obj.convertToSymmetricProperty();
            objectInformation.addSymmetricProp();
            if(inverse != null){
                inverse.convertToSymmetricProperty();
                objectInformation.addSymmetricProp();
            }
        }
        
        
        //check whether functionality is inherited
        if(isFunctional != 0){
            OntProperty superProp = obj.getSuperProperty();
            while(isFunctional != 0 && superProp != null){
                if(superProp.isFunctionalProperty())
                    isFunctional = 0;
                superProp = superProp.getSuperProperty();
            }
        }
        if(isFunctional == 0){
            //check whether symmetricity is inherited
            if(isSymmetric != 0){
                OntProperty superProp = obj.getSuperProperty();
                while(isSymmetric != 0 && superProp != null){
                    if(superProp.isSymmetricProperty())
                        isSymmetric = 0;
                    superProp = superProp.getSuperProperty();
                }
            }
            if(isSymmetric == 0){
                if(inverse != null){
                    inverse.convertToFunctionalProperty();
                    objectInformation.addFunctionalProp();
                    inverse.convertToInverseFunctionalProperty();
                }
                obj.convertToInverseFunctionalProperty();
            }
        }
    }
    
    /**
     * whether an object property is transitive.
     *
     * @param ObjectProperty obj: the object property of which we 
     *                            want to the characteristics.
     * @param ObjectProperty inverse: the inverse object property that 
     *                                needs characteristics that correspond
     *                                with the property of which it's the inverse.
     */
    private void addTransitive(ObjectProperty obj, ObjectProperty inverse){
        //is this a transitive property?
        //It must have the same domain and range
        ExtendedIterator it = obj.listDomain();
        OntoList domains = new OntoList();
        domains.addAll(it.toList());
        
        it = obj.listRange();
        OntoList ranges = new OntoList();
        ranges.addAll(it.toList());
        
        boolean equal = true;
        for(int i = 0; i < domains.size() && equal; i++){
            OntClass domain = domains.getOntClass(i);
            boolean found = false;
            for(int j = 0; j < ranges.size() && !found; j++){
                OntClass range = ranges.getOntClass(j);
                if(domain.equals(range)){
                    ranges.remove(range);
                    found = true;
                }
            }
            if(!found)
                equal = false;
        }
        
        if(equal){
            int isTransitive = pseudoRand.nextInt(2);
            //0 = transitive, 1 = not transitive
            //A property can't be transitive and functional
            boolean functional = false;
            OntProperty superProp = obj;
            while(superProp != null && !functional){
                functional = superProp.isFunctionalProperty();
                superProp = superProp.getSuperProperty();
            }
            
            if(!functional && isTransitive == 0){    
                obj.convertToTransitiveProperty();
                objectInformation.addTransitiveProp();
                if(inverse != null){
                    inverse.convertToTransitiveProperty();
                    objectInformation.addTransitiveProp();
                }
            }
        }
    }
    
    /**
     * Searches all the classes of which the connectivity demands are not fulfilled.  
     *
     * @param HashMap<OntClass,Integer> connect: Maps each class on his connectivity (= the
     *                                           number of classes he's connected to).
     * @return ArrayList<OntClass> : Contains all the classes of which the connectivity demands are
     *                               not fullfilled.
     */
    private ArrayList<OntClass> searchNotConnected(HashMap<OntClass,Integer> connect){
        Set<OntClass> keys = connect.keySet();
        Iterator<OntClass> it = keys.iterator();
        ArrayList<OntClass> notConnected = new ArrayList<OntClass>();
        while(it.hasNext())
        {
            OntClass ontClass = it.next();
            Integer connection = connect.get(ontClass);
            if(connection.intValue() < minConnectivity)
                notConnected.add(ontClass);
        }
        return notConnected;
    }
    
    /**
     * Create object properties for the enumerated classes that
     * will be generated in the future.
     *
     * @param int objCounter: the number of object properties that already 
     *                        have been created.
     * @return int: the number of object properties that already have been created.
     */
    public int createEnumObjProp(int objCounter){
        int number = nrOfEnumClasses * minConnectivity;
        for(int i = 0; i < number; i++){
            ObjectProperty obj = model.createObjectProperty(baseURI + "ObjectProperty" + objCounter);
            //Transitivity cannot be set yet as the object property doesn't have a domain and a range yet
            addFuncSymm(obj, null);
            objCounter++;
        }
        return objCounter;
    }
    
    /**
     * Generate the object properties (relations) that make sure that the
     * enumerated classes also satisfy the minimum connectivity constraints.
     */
    public void connectEnumClasses(){
        int objCounter = objectInformation.getNrOfProps();
        int numberOfObjProps = nrOfEnumClasses * minConnectivity;
        for(int i = 0; i < nrOfEnumClasses; i++){
            OntClass enumer = model.getOntClass(baseURI + "enumeratedClass" + i);
            connect.put(enumer,new Integer(0));
        }
        ArrayList<OntClass> notConnected = searchNotConnected(connect);
        
        for(int i = 1; i <= numberOfObjProps; i++){
            
            ObjectProperty objProp = model.getObjectProperty(baseURI + "ObjectProperty" + (objCounter - i));
            
            OntClass chosen = null;
            if(notConnected.size() > 0){
                //Randomly choose a class that doesn't have minimum connectivity yet
                int choose = pseudoRand.nextInt(notConnected.size());
                chosen = notConnected.get(choose);
            }
            else{
                //Randomly choose a class
                int choose = pseudoRand.nextInt(connectingClasses.size());
                chosen = connectingClasses.get(choose);
            }
            
            //Is it going to be the domain or the range?
            int domRange = pseudoRand.nextInt(2);
            //0: domain, 1:range
            if(domRange == 0)
                objProp.setDomain(chosen);  
            else
                objProp.setRange(chosen);
            
            //Search another random class to be the domain/range that isn't allready
            //the range/domain (because loops do not count towards the connectivity).
            OntClass other = null;
            while(other == chosen || other == null){
                int chooseOther = pseudoRand.nextInt(connectingClasses.size());
                other = connectingClasses.get(chooseOther);
            }
            if(domRange == 0)
                objProp.setRange(other);
            else
                objProp.setDomain(other);
            
            //adjust the connected table
            Integer oldConnect = connect.get(chosen);
            Integer newConnect = new Integer(oldConnect.intValue() + 1);
            connect.put(chosen,newConnect); 
            if(newConnect >= minConnectivity && notConnected.contains(chosen))
                notConnected.remove(chosen);
            
            oldConnect = connect.get(other);
            newConnect = new Integer(oldConnect.intValue() + 1);
            connect.put(other,newConnect); 
            //Check if the connectivity demands for this class are reached.
            if(newConnect >= minConnectivity && notConnected.contains(other))
                notConnected.remove(other);
            
            //Add if this property is transitive or not
            //This may not be done if this property is used 
            //to form a restriction class.
            if(!restrictionProps.contains(objProp))
                addTransitive(objProp, null);
        }
        
        objectInformation.setConnect(connect);
    }
    
    /////-------------------------Getters/Setters-------------------------/////
    
    /**
     * Get some feedback about the generated object property hierarchie of the ontology.
     * E.g. the number of relations, the number of subrelations and so on.
     *
     * @see ObjectInformation
     * @return ObjectInformation: an object containing all the information about the
     *                            generated object properties.
     */
    public ObjectInformation getObjectInformation(){
        return objectInformation;
    }

    /**
     * Sets the object properties used to create the restriction classes.
     *
     * @param ArrayList<OntProperty> restrictionProps: the object properties 
     *                                                 used to create the 
     *                                                 restriction classes.
     */
    public void setRestrictionProps(ArrayList<OntProperty> restrictionProps) {
        this.restrictionProps = restrictionProps;
    }
}
