/*
 * Copyright 2008-2011 Femke Ongenae, Stijn Verstichel
 * distributed under the terms of the AGPL License
 * 
 * This file is part of OTAGen.
 * 
 * OTAGen is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Femke Ongenae & Stijn Verstichel
 * Ghent University - INTEC - IBCN
 * Zuiderpoort Block C0 bus 201
 * Gaston Crommenlaan 8
 * B-9050 Ghent
 * Belgium
 * E-mail: Femke.Ongenae@intec.ugent.be
 * E-mail: Stijn.Verstichel@intec.ugent.be
 * 
 */
package otagen;

import com.hp.hpl.jena.ontology.DatatypeProperty;
import com.hp.hpl.jena.ontology.ObjectProperty;
import com.hp.hpl.jena.ontology.OntClass;
import com.hp.hpl.jena.ontology.OntModel;
import com.hp.hpl.jena.ontology.OntProperty;
import com.hp.hpl.jena.ontology.OntResource;
import java.util.List;
import java.util.Random;
import otagen.datatypes.OntoList;
import otagen.feedback.QueryInformation;

/**
 * This class generated the queries on the given OntModel according to the given
 * parameters. The same Random generator is used as for the other generating
 * procedures.
 * 
 * @author Femke Ongenae.
 * @author Stijn Verstichel.
 */
public class QueryGen {

    private boolean debug = false;
    //The base URI of the ontology
    private static String baseURI = "http://localhost/owl/ontologies/generated#";
    //a pseudo random generator (with seed);
    private Random pseudoRand = null;
    //holds the information about the generated A-Box.
    private QueryInformation feedBack = null;
    //specifies for every query-set to be generated the depths of the queries
    private int[] queryDepths = null;
    //specifies for every query-set the be generated the amount of the queries
    private int[] queryCounts = null;
    //Holds the Jena OntModel
    private OntModel model = null;
    //String used for concatenating the different pieces of the generated query
    private String tempQuery = null;
    //int counting the amount of variables used in the query
    private int variableCount = 0;

    /**
     * Generates the queries for the given model according to the specified
     * parameters
     * 
     * @param model: : The Jena internal OntModel, as constructed with
     * the generated T-Box
     * @param pseudoRand: The seed-based random generator, the same as
     * being used by the T-Box generator.
     * @param queryDepths: specifies for every query-set to be generated the 
     * depths of the queries
     * @param queryCounts: specifies for every query-set the be generated the 
     * amount of the queries
     */
    public QueryGen(OntModel model, Random pseudoRand, int[] queryDepths,
            int[] queryCounts) {
        //initialize the fields
        this.model = model;
        this.pseudoRand = pseudoRand;
        this.queryDepths = queryDepths;
        this.queryCounts = queryCounts;

        //create an empty QueryInformation object that will hold the feedback 
        //about the generated Queries
        this.feedBack = new QueryInformation();
        this.tempQuery = "";

        //business-logic
        this.generateQueries();
    }

    /**
     * Coordinating method for the Query-Generation process
     */
    private void generateQueries() {
        OntClass currentOntClass = null;
        int type = 0;
        List properties = null;
        OntProperty property = null;
        ObjectProperty objectProperty = null;
        DatatypeProperty datatypeProperty = null;
        OntResource range = null;
        List superClasses = null;
        OntClass superClass = null;
        List subClasses = null;
        OntClass subClass = null;
        for (int i = 0; i < queryDepths.length; i++) {
            for (int j = 0; j < queryCounts[i]; j++) {
                currentOntClass = this.chooseStartingOntClass();
                this.createQueryStart(currentOntClass);

                for (int k = 1; k < queryDepths[i]; k++) {
                    type = pseudoRand.nextInt(3);
                    type++;
                    //OntProperties
                    if (type == 0) {
                        if (currentOntClass != null) {
                            if (debug) {
                                System.out.println("[GENERATEQUERIES][DEBUG] type == 0");
                            }
                            properties = currentOntClass.listDeclaredProperties(false).toList();
                            if (properties.size() > 0) {
                                property = (OntProperty) properties.get(pseudoRand.nextInt(properties.size()));
                                if (property.isObjectProperty()) {
                                    objectProperty = property.asObjectProperty();
                                    if (objectProperty.getRange() != null) {
                                        this.createObjectPropertyQueryPart(objectProperty);
                                        range = objectProperty.getRange();
                                        currentOntClass = this.nextOntClass(range);
                                    } else {
                                        k--;
                                    }
                                } else if (property.isDatatypeProperty()) {
                                    datatypeProperty = property.asDatatypeProperty();
                                    this.createDatatypePropertyQueryPart(datatypeProperty);
                                }
                            } else {
                                k--;
                            }
                        }
                    } //SuperClass
                    else if (type == 1) {
                        if (currentOntClass != null && currentOntClass.listSuperClasses().toList().size() > 0) {
                            if (debug) {
                                System.out.println("[GENERATEQUERIES][DEBUG] type == 1");
                            }
                            superClasses = currentOntClass.listSuperClasses().toList();
                            superClass = (OntClass) superClasses.get(pseudoRand.nextInt(superClasses.size()));
                            this.createSuperClassQueryPart(currentOntClass, superClass);
                            currentOntClass = superClass;
                            k++;
                        } else {
                            k--;
                        }
                    } //SubClass
                    else if (type == 2) {
                        if (currentOntClass != null && currentOntClass.listSubClasses().toList().size() > 0) {
                            if (debug) {
                                System.out.println("[GENERATEDQUERIES][DEBUG] type == 2");
                            }
                            subClasses = currentOntClass.listSubClasses().toList();
                            subClass = (OntClass) subClasses.get(pseudoRand.nextInt(subClasses.size()));
                            this.createSubClassQueryPart(subClass, currentOntClass);
                            currentOntClass = subClass;
                            k++;
                        } else {
                            k--;
                        }
                    }
                }
                this.completeQuery(tempQuery);
                if (!feedBack.addQuery(tempQuery)) {
                    //System.out.println("notAdded:" + j + "in set: " + i);
                    j--;
                }
                tempQuery = "";
                variableCount = 0;
            }
        }
    //System.out.println(this.tempQuery);
    }

    /**
     * Election method for the starting OntClass concept
     * 
     * @return OntClass ontClass: Starting point for the query
     */
    private OntClass chooseStartingOntClass() {
        OntoList ontoList = new OntoList();
        ontoList.addAll(model.listNamedClasses().toList());
        int index = pseudoRand.nextInt(ontoList.size());
        if (debug) {
            System.out.println("[CHOOSESTARTINGONTCLASS][DEBUG] Starting OntClass: " +
                    ontoList.getOntClass(index));
        }
        while (ontoList.getOntClass(index) == null) {
            index = pseudoRand.nextInt(ontoList.size());
        }
        return ontoList.getOntClass(index);
    }

    /**
     * Creates the first part of the query as this part is a special case
     * 
     * @param ontClass ontclass
     */
    private void createQueryStart(OntClass ontClass) {
        this.tempQuery = this.tempQuery +
                "?x" + variableCount + " rdf:type <" + ontClass.getURI() + ">";
        variableCount++;
    }

    /**
     * Creates a query part following the path of an ObjectProperty
     * 
     * @param objectProperty
     */
    private void createObjectPropertyQueryPart(ObjectProperty objectProperty) {
        this.tempQuery = this.tempQuery +
                " . \n?x" + (variableCount - 1) + " <" + objectProperty.getURI() + "> " +
                "?x" + variableCount;
        variableCount++;
    }

    /**
     * Creates a query part following the path of an DatatypeProperty
     * 
     * @param datatypeProperty
     */
    private void createDatatypePropertyQueryPart(DatatypeProperty datatypeProperty) {
        this.tempQuery = this.tempQuery +
                " . \n?x" + (variableCount - 1) + " <" + datatypeProperty.getURI() + "> " +
                "?x" + variableCount;
        variableCount++;
    }

    /**
     * Creates a query part following the superClass subsumption
     * 
     * @param subClass
     * @param superClass
     */
    private void createSuperClassQueryPart(OntClass subClass, OntClass superClass) {
        this.tempQuery = this.tempQuery +
                " . \n<" + subClass.getURI() + "> rdfs:subClassOf <" +
                superClass.getURI() + ">" +
                " . \n?x" + variableCount + " rdf:type <" + superClass.getURI() + ">";
        variableCount++;
    }

    /**
     * Creates a query part following the subClass subsumption
     * 
     * @param subClass
     * @param superClass
     */
    private void createSubClassQueryPart(OntClass subClass, OntClass superClass) {
        this.tempQuery = this.tempQuery +
                " . \n<" + subClass.getURI() + "> rdfs:subClassOf <" +
                superClass.getURI() + ">" +
                " . \n?x" + variableCount + " rdf:type <" + subClass.getURI() + ">";
        variableCount++;
    }

    /**
     * This method adds the prefixes, the select and the where
     * keywords
     * 
     * @param query
     */
    private void completeQuery(String query) {
        String prefix = "PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>\n";
        prefix += "PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>\n";
        prefix += "PREFIX owl: <http://www.w3.org/2002/07/owl#>\n";

        String select = " SELECT ";
        for (int i = 0; i < variableCount; i++) {
            select = select + "?x" + i + " ";
        }
        String startWhere = " WHERE { ";
        String endWhere = " } ";

        tempQuery = prefix + select + startWhere + tempQuery + endWhere;
    }

    /**
     * In case an ObjectProperty track is followed, return either the direct
     * OntClass in case a simple range is used, or otherwise an arbitrary
     * OntClass from the complex range expression
     * 
     * @param resource range
     * @return OntClass ontclass
     */
    private OntClass nextOntClass(OntResource resource) {
        if (resource.isURIResource()) {
            return model.getOntClass(resource.getURI());
        } else if (resource.asClass().isEnumeratedClass()) {
            return (OntClass) resource.asClass().asEnumeratedClass().listOneOf().toList().get(resource.asClass().asEnumeratedClass().listOneOf().toList().size());
        } else if (resource.asClass().isUnionClass()) {
            return (OntClass) resource.asClass().asUnionClass().listOperands().toList().get(resource.asClass().asUnionClass().listOperands().toList().size());
        } else {
            return null;
        }
    }

    public QueryInformation getFeedBack() {
        return feedBack;
    }
}
