/*
 * Copyright 2008-2011 Femke Ongenae, Stijn Verstichel
 * distributed under the terms of the AGPL License
 * 
 * This file is part of OTAGen.
 * 
 * OTAGen is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Femke Ongenae & Stijn Verstichel
 * Ghent University - INTEC - IBCN
 * Zuiderpoort Block C0 bus 201
 * Gaston Crommenlaan 8
 * B-9050 Ghent
 * Belgium
 * E-mail: Femke.Ongenae@intec.ugent.be
 * E-mail: Stijn.Verstichel@intec.ugent.be
 * 
 */

package otagen;

import com.hp.hpl.jena.ontology.OntModel;
import java.util.Random;
import otagen.feedback.TboxInformation;


/**
 * Generates the conceptual level of the ontology (the class hierarchie,
 * the object properties and the datatype properties).
 *
 * @author Femke Ongenae.
 * @author Stijn Verstichel.
 */
public class TboxGen {

    /////-------------------------FIELDS-------------------------/////
    
    //The base URI of the ontology
    private static String baseURI = "http://localhost/owl/ontologies/generated#";
    
    //holds the information about the generated T-Box.
    private TboxInformation feedBack = null;
    
    //Used to generate the class hierarchie of the ontology
    private ClassesGen classesGen = null;
    
    //Used to generate the object property hierarchie (relations)
    //of the ontology.
    private ObjPropsGen objPropsGen = null;
    
    /////-------------------------Constructor-------------------------/////
    
    /**
     * Generates the conceptual level of the ontology (T-Box).
     *
     * @param int nrOfClasses: the number of classes.
     * @param int maxDepth: the maximal depth of the classes and properties.
     * @param int maxNrSubClasses: the maximal number of subclasses a class can have.
     * @param boolean flag: A flag: true : create logical defined classes,
     *                              false : create some empty (non-defined) classes.
     * @param int nrOfFlagClasses: the number of classes of the previous category that must be created.
     * @param int minConnectivity: the minimal connectivity of class-relations graph.
     * @param int maxNrPercProperties: the maximal percentage of object properties (relations) or 
     *                                  data properties with complicated domain/range specifications.
     * @param int maxNrSubProperties: the maximal number of subproperties a property can have.
     * @param OntModel model: the ontology we're creating.
     * @param Random pseudoRand: a pseudo random generator (with seed).
     */
    public TboxGen(int nrOfClasses, int maxDepth, int maxNrSubClasses, boolean flag, int nrOfFlagClasses,
                   int minConnectivity, int maxPercComplProperties, int maxNrSubProperties,
                   OntModel model, Random pseudoRand) {
        
        //create an empty Tboxinformation object that will hold the feedback about this generated T-Box
        feedBack = new TboxInformation();
        
        //create the class structure of the ontology (classes + hierarchie)
        System.out.println("[STATUS] Generating Classes");
        classesGen = new ClassesGen(model, baseURI, nrOfClasses, pseudoRand, maxDepth, 
                                    flag, nrOfFlagClasses, maxNrSubClasses);
        //configure feedback
        feedBack.setClassInformation(classesGen.getClassInformation());
        
        //generate the datatype properties
        System.out.println("[STATUS] Generating Datatype Properties");
        DataPropsGen dataPropsGen = new DataPropsGen(model, baseURI, pseudoRand, maxDepth,
                                                     maxPercComplProperties, maxNrSubProperties,
                                                     classesGen.getConnectingClasses());
        //configure feedback
        feedBack.setDatatypeInformation(dataPropsGen.getDatatypeInformation());

        //generate a graph (the relations)
        System.out.println("[STATUS] Generating Relations (Object Properties)");
        objPropsGen = new ObjPropsGen(model, baseURI, pseudoRand, maxDepth, minConnectivity, 
                                      maxPercComplProperties, maxNrSubProperties,
                                      feedBack.getClassInformation().getNrOfEnumClasses(),
                                      feedBack.getClassInformation().getNrOfRestrictionClasses(),
                                      classesGen.getConnectingClasses());
        //configure feedback
        feedBack.setObjectInformation(objPropsGen.getObjectInformation());
        
        //Generate restriction classes
        System.out.println("[STATUS] Generating Restriction Classes");
        classesGen.generateRestrictionClasses();
        objPropsGen.setRestrictionProps(classesGen.getRestrictionProps());
    }
    
    /**
     * Generate enumerated classes (classes that only consist of the specified
     * individuals).
     */
    public void generateEnumClasses(){
        classesGen.createEnumeratedClasses();
    }
    
    /**
     * Generate the object properties (relations) that make sure that the
     * enumerated classes also satisfy the minimum connectivity constraints.
     */
    public void connectEnumClasses(){
        objPropsGen.connectEnumClasses();
    }
    
    
    /////-------------------------Getters/Setters-------------------------/////
    
    /**
     * Get some feedback about the generated T-Box (conceptual level) of the ontology.
     * E.g. the number of relations, the number of datatype properties and so on.
     * @see TboxInformation
     * @return TBoxInformation: an object containing all the information about the
     *                          generated T-Box.
     */
    public TboxInformation getFeedBack(){
        return feedBack;
    }
}
