/*
 * Copyright 2008-2011 Femke Ongenae, Stijn Verstichel
 * distributed under the terms of the AGPL License
 * 
 * This file is part of OTAGen.
 * 
 * OTAGen is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Femke Ongenae & Stijn Verstichel
 * Ghent University - INTEC - IBCN
 * Zuiderpoort Block C0 bus 201
 * Gaston Crommenlaan 8
 * B-9050 Ghent
 * Belgium
 * E-mail: Femke.Ongenae@intec.ugent.be
 * E-mail: Stijn.Verstichel@intec.ugent.be
 * 
 */
package otagen.datatypes;

import com.hp.hpl.jena.ontology.OntClass;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Random;

/**
 * Session object containing all variables needed for the individual
 * generating process
 * 
 * @author Femke Ongenae.
 * @author Stijn Verstichel.
 */
public class IndividualGeneratorSessionVariables {

    //Contains the mapping of all concepts to be instantiated with the
    //respective currently generated individual count.
    private Map<OntClass, Integer> generatedCountMap = null;
    //Specifies the minimum individuals to be generated for each concept
    private int minGeneratedCount = 0;
    //Reference to the Pseudo-Random generator (according to the seed)
    private Random pseudoRandom = null;

    public IndividualGeneratorSessionVariables(int minGeneratedCount,
            Random pseudoRandom, List<OntClass> classes) {
        this.minGeneratedCount = minGeneratedCount;
        this.pseudoRandom = pseudoRandom;
        generatedCountMap = new HashMap<OntClass, Integer>();
        Iterator<OntClass> it = classes.iterator();
        while (it.hasNext()) {
            generatedCountMap.put(it.next(), new Integer(0));
        }
    }

    public Map<OntClass, Integer> getGeneratedCountMap() {
        return generatedCountMap;
    }

    public void setGeneratedCountMap(Map<OntClass, Integer> generatedCountMap) {
        this.generatedCountMap = generatedCountMap;
    }

    /**
     * Returns the next concept to be instantiated and augments
     * the corresponding value in the Map.
     * 
     * @return OntClass
     */
    public OntClass getNextOntClass() {
        if (this.thresholdReached()) {
            return null;
        } else {
            int index = pseudoRandom.nextInt(generatedCountMap.size());
            OntClass ontClass = (OntClass) generatedCountMap.keySet().toArray()[index];
            Integer currentValue = generatedCountMap.get(ontClass);
            Integer newValue = new Integer(currentValue.intValue() + 1);
            generatedCountMap.remove(ontClass);
            generatedCountMap.put(ontClass, newValue);
            return ontClass;
        }
    }

    /**
     * Checks whether all concepts satisfy the minimum amount of
     * individuals
     * 
     * @return boolean
     */
    private boolean thresholdReached() {
        Collection<Integer> values = generatedCountMap.values();
        Iterator<Integer> it = values.iterator();
        while (it.hasNext()) {
            if (it.next().intValue() < minGeneratedCount) {
                return false;
            }
        }
        return true;
    }
}
