/*
 * Copyright 2008-2011 Femke Ongenae, Stijn Verstichel
 * distributed under the terms of the AGPL License
 * 
 * This file is part of OTAGen.
 * 
 * OTAGen is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Femke Ongenae & Stijn Verstichel
 * Ghent University - INTEC - IBCN
 * Zuiderpoort Block C0 bus 201
 * Gaston Crommenlaan 8
 * B-9050 Ghent
 * Belgium
 * E-mail: Femke.Ongenae@intec.ugent.be
 * E-mail: Stijn.Verstichel@intec.ugent.be
 * 
 */

package otagen.feedback;

/**
 * Class used to give feedback information about the creation of the datatype
 * properties of the ontology eg. the number of properties and so on.
 *
 * @author Femke Ongenae.
 * @author Stijn Verstichel.
 */
public class DatatypeInformation {
    
    /////-------------------------FIELDS-------------------------/////
    
    //Counts the number of datatype properties that have been created.
    private int nrOfProps = 0;
    
    //Counts the number of datatype properties with complicated domains 
    //(a domain that consists of more than one class) that have been created.
    private int nrOfComplProps = 0;
    
    //Counts the number of functional datatype properties 
    //that have been created.
    private int nrOfFunctionalProps = 0;
    
    //Counts the maximum depth of datatype properties tree.
    private int maxDepthProps = 0;
    
    //Counts the number of datatype properties with range String
    //that have been created.
    private int nrOfRangeStringProps = 0;
    
    //Counts the number of datatype properties with range integer
    //that have been created.
    private int nrOfRangeIntProps = 0;
    
    /////-------------------------Constructor-------------------------/////
    
    /** Creates a new instance of DatatypeInformation */
    public DatatypeInformation() {
    }
    
    /////-------------------------Business methods-------------------------/////
    
    /**
     * Raises the number of datatype properties that have been created with
     * range String with one.
     */
    public void addRangeStringDataProp(){
        nrOfRangeStringProps++;
    }
    
    /**
     * Raises the number of datatype properties that have been created with
     * range integer with one.
     */
    public void addRangeIntegerDataProp(){
        nrOfRangeIntProps++;
    }
    
    /**
     * Raises the number of functional datatype properties 
     * that have been created with one.
     */
    public void addFunctionalDataProp(){
        nrOfFunctionalProps++;
    }

    /////-------------------------Getters/Setters-------------------------/////
    
    /**
     * Gets the number of datatype properties that have been created.
     *
     * @return int: the number of datatype properties that have been created.
     */
    public int getNrOfProps() {
        return nrOfProps;
    }

    /**
     * Sets the number of datatype properties that have been created.
     *
     * @param int nrOfProps: the number of datatype properties 
     *                       that have been created.
     */
    public void setNrOfProps(int nrOfProps) {
        this.nrOfProps = nrOfProps;
    }

    /**
     * Gets the number of datatype properties with a complicated domain
     * (consisting of more than one class) that have been created.
     *
     * @return int: the number of complicated datatype properties 
     *              that have been created.
     */
    public int getNrOfComplProps() {
        return nrOfComplProps;
    }

    /**
     * Sets the number of datatype properties with a complicated domain
     * (consisting of more than one class) that have been created.
     *
     * @param int nrOfComplProps: the number of complicated datatype properties 
     *                            that have been created.
     */
    public void setNrOfComplProps(int nrOfComplProps) {
        this.nrOfComplProps = nrOfComplProps;
    }

    /**
     * Gets the number of functional datatype properties that have been created.
     *
     * @return int: the number of functional datatype 
     *              properties that have been created.
     */
    public int getNrOfFunctionalProps() {
        return nrOfFunctionalProps;
    } 
    
    /**
     * Sets the number of functional datatype properties that have been created.
     *
     * @param int nrOfFunctionalProps: the number of functional datatype 
     *                                 properties that have been created.
     */
    public void setNrOfFunctionalProps(int nrOfFunctionalProps) {
        this.nrOfFunctionalProps = nrOfFunctionalProps;
    }

    /**
     * Gets the maximal depth op the datatype properties tree.
     *
     * @return int: the maximal depth op the tree.
     */
    public int getMaxDepthProps() {
        return maxDepthProps;
    }

    /**
     * Sets the maximal depth op the datatype properties tree.
     *
     * @param int maxDepthProps: the maximal depth op the tree.
     */
    public void setMaxDepthProps(int maxDepthProps) {
        this.maxDepthProps = maxDepthProps;
    }
    
    /**
     * Gets the number of datatype properties that have been created with
     * range String.
     *
     * @return int: the number of datatype properties that have been created
     *              with range String.
     */
    public int getNrOfRangeStringProps() {
        return nrOfRangeStringProps;
    }

    /**
     * Sets the number of datatype properties that have been created with
     * range String.
     *
     * @param int nrOfRangeStringProps: the number of datatype properties 
     *                                  that have been created
     *                                  with range String.
     */
    public void setNrOfRangeStringProps(int nrOfRangeStringProps) {
        this.nrOfRangeStringProps = nrOfRangeStringProps;
    }

    /**
     * Gets the number of datatype properties that have been created with
     * range integer.
     *
     * @return int: the number of datatype properties that have been created
     *              with range integer.
     */
    public int getNrOfRangeIntProps() {
        return nrOfRangeIntProps;
    }

    /**
     * Sets the number of datatype properties that have been created with
     * range integer.
     *
     * @param int nrOfRangeIntProps: the number of datatype properties 
     *                               that have been created
     *                               with range integer.
     */
    public void setNrOfRangeIntProps(int nrOfRangeIntProps) {
        this.nrOfRangeIntProps = nrOfRangeIntProps;
    } 
   
}
