/*
 * Copyright 2008-2011 Femke Ongenae, Stijn Verstichel
 * distributed under the terms of the AGPL License
 * 
 * This file is part of OTAGen.
 * 
 * OTAGen is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Femke Ongenae & Stijn Verstichel
 * Ghent University - INTEC - IBCN
 * Zuiderpoort Block C0 bus 201
 * Gaston Crommenlaan 8
 * B-9050 Ghent
 * Belgium
 * E-mail: Femke.Ongenae@intec.ugent.be
 * E-mail: Stijn.Verstichel@intec.ugent.be
 * 
 */

package otagen.feedback;

import com.hp.hpl.jena.ontology.OntClass;
import java.util.HashMap;

/**
 * Class used to give feedback information about the creation of the object
 * properties of the ontology eg. the number of properties and so on.
 *
 * @author Femke Ongenae.
 * @author Stijn Verstichel.
 */
public class ObjectInformation {
    
    /////-------------------------FIELDS-------------------------/////
    
    //Counts the number of object properties that have been created.
    private int nrOfProps = 0;
    
    //Counts the maximum depth of object properties tree.
    private int maxDepthProps = 0;
    
    //Counts the number of functional object properties that have been created.
    private int nrOfFunctionalProps = 0;
    
    //Counts the number of symmetric object properties that have been created.
    private int nrOfSymmetricProps = 0;
    
    //Counts the number of transitive object properties that have been created.
    private int nrOfTransitiveProps = 0;
    
    //Counts the number of object properties with complicated domains/ranges
    //(a domain/range that consists of more than one class) 
    //that have been created.
    private int nrOfComplProps = 0;
    
    //Counts the number of object properties with complicated domains
    //(a domain that consists of more than one class) that have been created.
    private int nrOfComplDomProps = 0;
   
    //Counts the number of object properties with complicated ranges
    //(a range that consists of more than one class) that have been created.
    private int nrOfComplRangeProps = 0;
    
    //Counts the number of inverse properties that have been created.
    private int nrOfInverseProps = 0;
    
    //Count the number of times a class is connected with another class
    private HashMap<OntClass,Integer> connect = null;
    
    /////-------------------------Constructor-------------------------/////
    
    /** Creates a new instance of ObjectInformation */
    public ObjectInformation() {
    }
    
    /////-------------------------Business methods-------------------------/////
    
    /**
     * Raises the number of functional object properties 
     * that have been created with complicated domains with one.
     */
    public void addComplDomProp(){
        nrOfComplDomProps++;
    }
    
    /**
     * Raises the number of functional object properties 
     * that have been created with complicated ranges with one.
     */
    public void addComplRangeProp(){
        nrOfComplRangeProps++;
    }
    
    /**
     * Raises the number of functional object properties 
     * that have been created with one.
     */
    public void addFunctionalProp(){
        nrOfFunctionalProps++;
    }
    
    /**
     * Raises the number of symmetric object properties 
     * that have been created with one.
     */
    public void addSymmetricProp(){
        nrOfSymmetricProps++;
    }
    
    /**
     * Raises the number of transitive object properties 
     * that have been created with one.
     */
    public void addTransitiveProp(){
        nrOfTransitiveProps++;
    }
    
    /**
     * Raises the number of inverse object properties
     * that have been created with one.
     */
    public void addInverseProp() {
        nrOfInverseProps++;
    }

    /////-------------------------Getters/Setters-------------------------/////
    
    /**
     * Gets the number of object properties that have been created.
     *
     * @return int: the number of object properties that have been created.
     */
    public int getNrOfProps() {
        return nrOfProps;
    }

    /**
     * Sets the number of object properties that have been created.
     *
     * @param int nrOfProps: the number of object properties 
     *                       that have been created.
     */
    public void setNrOfProps(int nrOfProps) {
        this.nrOfProps = nrOfProps;
    }

    /**
     * Gets the maximal depth op the object properties tree.
     *
     * @return int: the maximal depth op the tree.
     */
    public int getMaxDepthProps() {
        return maxDepthProps;
    }

    /**
     * Sets the maximal depth op the object properties tree.
     *
     * @param int maxDepthProps: the maximal depth op the tree.
     */
    public void setMaxDepthProps(int maxDepthProps) {
        this.maxDepthProps = maxDepthProps;
    }

    /**
     * Gets the number of functional object properties that have been created.
     *
     * @return int: the number of functional object 
     *              properties that have been created.
     */
    public int getNrOfFunctionalProps() {
        return nrOfFunctionalProps;
    }

    /**
     * Sets the number of functional object properties that have been created.
     *
     * @param int nrOfFunctionalProps: the number of functional object 
     *                                 properties that have been created.
     */
    public void setNrOfFunctionalProps(int nrOfFunctionalProps) {
        this.nrOfFunctionalProps = nrOfFunctionalProps;
    }

    /**
     * Gets the number of symmetric object properties that have been created.
     *
     * @return int: the number of symmetric object 
     *              properties that have been created.
     */
    public int getNrOfSymmetricProps() {
        return nrOfSymmetricProps;
    }

    /**
     * Sets the number of symmetric object properties that have been created.
     *
     * @param int nrOfSymmetricProps: the number of symmetric object 
     *                                 properties that have been created.
     */
    public void setNrOfSymmetricProps(int nrOfSymmetricProps) {
        this.nrOfSymmetricProps = nrOfSymmetricProps;
    }

    /**
     * Gets the number of transitive object properties that have been created.
     *
     * @return int: the number of transitive object 
     *              properties that have been created.
     */
    public int getNrOfTransitiveProps() {
        return nrOfTransitiveProps;
    }

     /**
     * Sets the number of transitive object properties that have been created.
     *
     * @param int nrOfTransitiveProps: the number of transitive object 
     *                                 properties that have been created.
     */
    public void setNrOfTransitiveProps(int nrOfTransitiveProps) {
        this.nrOfTransitiveProps = nrOfTransitiveProps;
    }

    /**
     * Gets the number of object properties with a complicated domain/range
     * (consisting of more than one class) that have been created.
     *
     * @return int: the number of complicated object properties 
     *              that have been created.
     */
    public int getNrOfComplProps() {
        return nrOfComplProps;
    }

    /**
     * Sets the number of object properties with a complicated domain/range
     * (consisting of more than one class) that have been created.
     *
     * @param int nrOfComplProps: the number of complicated object properties 
     *                            that have been created.
     */
    public void setNrOfComplProps(int nrOfComplProps) {
        this.nrOfComplProps = nrOfComplProps;
    }

    /**
     * Gets the number of object properties with a complicated domain
     * (consisting of more than one class) that have been created.
     *
     * @return int: the number of complicated domain object properties 
     *              that have been created.
     */
    public int getNrOfComplDomProps() {
        return nrOfComplDomProps;
    }

    /**
     * Sets the number of object properties with a complicated domain
     * (consisting of more than one class) that have been created.
     *
     * @param int nrOfComplDomProps: the number of complicated domain object 
     *                               properties that have been created.
     */
    public void setNrOfComplDomProps(int nrOfComplDomProps) {
        this.nrOfComplDomProps = nrOfComplDomProps;
    }

    /**
     * Gets the number of object properties with a complicated range
     * (consisting of more than one class) that have been created.
     *
     * @return int: the number of complicated range object properties 
     *              that have been created.
     */
    public int getNrOfComplRangeProps() {
        return nrOfComplRangeProps;
    }

    /**
     * Sets the number of object properties with a complicated range
     * (consisting of more than one class) that have been created.
     *
     * @param int nrOfComplRangeProps: the number of complicated range object 
     *                                 properties that have been created.
     */
    public void setNrOfComplRangeProps(int nrOfComplRangeProps) {
        this.nrOfComplRangeProps = nrOfComplRangeProps;
    }

    /**
     * Gets the number of inverse object properties.
     * 
     * @return int: the number of inverse properties.
     */
    public int getNrOfInverseProps() {
        return nrOfInverseProps;
    }

    /**
     * Sets the number of inverse object properties.
     *
     * @param int nrOfInverseProps: the number of inverse properties.
     */
    public void setNrOfInverseProps(int nrOfInverseProps) {
        this.nrOfInverseProps = nrOfInverseProps;
    }

    /**
     * Sets the number of times a class is connected with another class.
     *
     * @return HashMap<OntClass, Integer>: maps each class on its connectivity.
     */
    public HashMap<OntClass, Integer> getConnect() {
        return connect;
    }
    
    /**
     * Gets the number of times each class is connected with another class.
     *
     * @param HashMap<OntClass, Integer> connect: maps each class on its connectivity.
     */
    public void setConnect(HashMap<OntClass, Integer> connect) {
        this.connect = connect;
    }
    
}